#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from enum import Enum
import json
import os
import socket
from typing import IO, Any, Dict, List, Union, Optional, Tuple, Iterator, cast

from pyspark.serializers import write_int, read_int, UTF8Deserializer
from pyspark.sql.pandas.serializers import ArrowStreamSerializer
from pyspark.sql.types import (
    StructType,
    Row,
)
from pyspark.sql.pandas.types import convert_pandas_using_numpy_type
from pyspark.serializers import CPickleSerializer
from pyspark.errors import PySparkRuntimeError
import uuid

__all__ = ["StatefulProcessorApiClient", "StatefulProcessorHandleState"]


class StatefulProcessorHandleState(Enum):
    PRE_INIT = 0
    CREATED = 1
    INITIALIZED = 2
    DATA_PROCESSED = 3
    TIMER_PROCESSED = 4
    CLOSED = 5


class StatefulProcessorApiClient:
    def __init__(
        self, state_server_port: Union[int, str], key_schema: StructType, is_driver: bool = False
    ) -> None:
        self.key_schema = key_schema
        if isinstance(state_server_port, str):
            self._client_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            self._client_socket.connect(state_server_port)
        else:
            self._client_socket = socket.socket()
            self._client_socket.connect(("localhost", state_server_port))

            # SPARK-51667: We have a pattern of sending messages continuously from one side
            # (Python -> JVM, and vice versa) before getting response from other side. Since most
            # messages we are sending are small, this triggers the bad combination of Nagle's
            # algorithm and delayed ACKs, which can cause a significant delay on the latency.
            # See SPARK-51667 for more details on how this can be a problem.
            #
            # Disabling either would work, but it's more common to disable Nagle's algorithm; there
            # is lot less reference to disabling delayed ACKs, while there are lots of resources to
            # disable Nagle's algorithm.
            self._client_socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

        self.sockfile = self._client_socket.makefile(
            "rwb", int(os.environ.get("SPARK_BUFFER_SIZE", 65536))
        )
        if is_driver:
            self.handle_state = StatefulProcessorHandleState.PRE_INIT
        else:
            self.handle_state = StatefulProcessorHandleState.CREATED
        self.utf8_deserializer = UTF8Deserializer()
        self.pickleSer = CPickleSerializer()
        self.serializer = ArrowStreamSerializer()
        # Dictionaries to store the mapping between iterator id and a tuple of data batch
        # and the index of the last row that was read.
        self.list_timer_iterator_cursors: Dict[str, Tuple[Any, int, bool]] = {}
        self.expiry_timer_iterator_cursors: Dict[str, Tuple[Any, int, bool]] = {}

        # statefulProcessorApiClient is initialized per batch per partition,
        # so we will have new timestamps for a new batch
        self._batch_timestamp = -1
        self._watermark_timestamp = -1

    def set_handle_state(self, state: StatefulProcessorHandleState) -> None:
        import pyspark.sql.streaming.proto.StateMessage_pb2 as stateMessage

        if state == StatefulProcessorHandleState.PRE_INIT:
            proto_state = stateMessage.PRE_INIT
        elif state == StatefulProcessorHandleState.CREATED:
            proto_state = stateMessage.CREATED
        elif state == StatefulProcessorHandleState.INITIALIZED:
            proto_state = stateMessage.INITIALIZED
        elif state == StatefulProcessorHandleState.DATA_PROCESSED:
            proto_state = stateMessage.DATA_PROCESSED
        elif state == StatefulProcessorHandleState.TIMER_PROCESSED:
            proto_state = stateMessage.TIMER_PROCESSED
        else:
            proto_state = stateMessage.CLOSED
        set_handle_state = stateMessage.SetHandleState(state=proto_state)
        handle_call = stateMessage.StatefulProcessorCall(setHandleState=set_handle_state)
        message = stateMessage.StateRequest(statefulProcessorCall=handle_call)

        self._send_proto_message(message.SerializeToString())

        response_message = self._receive_proto_message()
        status = response_message[0]
        if status == 0:
            self.handle_state = state
        else:
            # TODO(SPARK-49233): Classify errors thrown by internal methods.
            raise PySparkRuntimeError(f"Error setting handle state: {response_message[1]}")

    def set_implicit_key(self, key: Tuple) -> None:
        import pyspark.sql.streaming.proto.StateMessage_pb2 as stateMessage

        key_bytes = self._serialize_to_bytes(self.key_schema, key)
        set_implicit_key = stateMessage.SetImplicitKey(key=key_bytes)
        request = stateMessage.ImplicitGroupingKeyRequest(setImplicitKey=set_implicit_key)
        message = stateMessage.StateRequest(implicitGroupingKeyRequest=request)

        self._send_proto_message(message.SerializeToString())
        response_message = self._receive_proto_message()
        status = response_message[0]
        if status != 0:
            # TODO(SPARK-49233): Classify errors thrown by internal methods.
            raise PySparkRuntimeError(f"Error setting implicit key: {response_message[1]}")

    def remove_implicit_key(self) -> None:
        import pyspark.sql.streaming.proto.StateMessage_pb2 as stateMessage

        remove_implicit_key = stateMessage.RemoveImplicitKey()
        request = stateMessage.ImplicitGroupingKeyRequest(removeImplicitKey=remove_implicit_key)
        message = stateMessage.StateRequest(implicitGroupingKeyRequest=request)

        self._send_proto_message(message.SerializeToString())
        response_message = self._receive_proto_message()
        status = response_message[0]
        if status != 0:
            # TODO(SPARK-49233): Classify errors thrown by internal methods.
            raise PySparkRuntimeError(f"Error removing implicit key: {response_message[1]}")

    def get_value_state(
        self, state_name: str, schema: Union[StructType, str], ttl_duration_ms: Optional[int]
    ) -> None:
        import pyspark.sql.streaming.proto.StateMessage_pb2 as stateMessage

        if isinstance(schema, str):
            schema = self._parse_string_schema(schema)

        state_call_command = stateMessage.StateCallCommand()
        state_call_command.stateName = state_name
        state_call_command.schema = schema.json()
        if ttl_duration_ms is not None:
            state_call_command.ttl.durationMs = ttl_duration_ms
        call = stateMessage.StatefulProcessorCall(getValueState=state_call_command)
        message = stateMessage.StateRequest(statefulProcessorCall=call)

        self._send_proto_message(message.SerializeToString())
        response_message = self._receive_proto_message()
        status = response_message[0]
        if status != 0:
            # TODO(SPARK-49233): Classify user facing errors.
            raise PySparkRuntimeError(f"Error initializing value state: {response_message[1]}")

    def get_list_state(
        self, state_name: str, schema: Union[StructType, str], ttl_duration_ms: Optional[int]
    ) -> None:
        import pyspark.sql.streaming.proto.StateMessage_pb2 as stateMessage

        if isinstance(schema, str):
            schema = self._parse_string_schema(schema)

        state_call_command = stateMessage.StateCallCommand()
        state_call_command.stateName = state_name
        state_call_command.schema = schema.json()
        if ttl_duration_ms is not None:
            state_call_command.ttl.durationMs = ttl_duration_ms
        call = stateMessage.StatefulProcessorCall(getListState=state_call_command)
        message = stateMessage.StateRequest(statefulProcessorCall=call)

        self._send_proto_message(message.SerializeToString())
        response_message = self._receive_proto_message()
        status = response_message[0]
        if status != 0:
            # TODO(SPARK-49233): Classify user facing errors.
            raise PySparkRuntimeError(f"Error initializing list state: {response_message[1]}")

    def register_timer(self, expiry_time_stamp_ms: int) -> None:
        import pyspark.sql.streaming.proto.StateMessage_pb2 as stateMessage

        register_call = stateMessage.RegisterTimer(expiryTimestampMs=expiry_time_stamp_ms)
        state_call_command = stateMessage.TimerStateCallCommand(register=register_call)
        call = stateMessage.StatefulProcessorCall(timerStateCall=state_call_command)
        message = stateMessage.StateRequest(statefulProcessorCall=call)

        self._send_proto_message(message.SerializeToString())
        response_message = self._receive_proto_message()
        status = response_message[0]
        if status != 0:
            # TODO(SPARK-49233): Classify user facing errors.
            raise PySparkRuntimeError(f"Error register timer: {response_message[1]}")

    def delete_timer(self, expiry_time_stamp_ms: int) -> None:
        import pyspark.sql.streaming.proto.StateMessage_pb2 as stateMessage

        delete_call = stateMessage.DeleteTimer(expiryTimestampMs=expiry_time_stamp_ms)
        state_call_command = stateMessage.TimerStateCallCommand(delete=delete_call)
        call = stateMessage.StatefulProcessorCall(timerStateCall=state_call_command)
        message = stateMessage.StateRequest(statefulProcessorCall=call)

        self._send_proto_message(message.SerializeToString())
        response_message = self._receive_proto_message()
        status = response_message[0]
        if status != 0:
            # TODO(SPARK-49233): Classify user facing errors.
            raise PySparkRuntimeError(f"Error deleting timer: {response_message[1]}")

    def get_list_timer_row(self, iterator_id: str) -> Tuple[int, bool]:
        import pyspark.sql.streaming.proto.StateMessage_pb2 as stateMessage

        if iterator_id in self.list_timer_iterator_cursors:
            # if the iterator is already in the dictionary, return the next row
            data_batch, index, require_next_fetch = self.list_timer_iterator_cursors[iterator_id]
        else:
            list_call = stateMessage.ListTimers(iteratorId=iterator_id)
            state_call_command = stateMessage.TimerStateCallCommand(list=list_call)
            call = stateMessage.StatefulProcessorCall(timerStateCall=state_call_command)
            message = stateMessage.StateRequest(statefulProcessorCall=call)

            self._send_proto_message(message.SerializeToString())
            response_message = self._receive_proto_message_with_timers()
            status = response_message[0]
            if status == 0:
                data_batch = list(map(lambda x: x.timestampMs, response_message[2]))
                require_next_fetch = response_message[3]
                index = 0
            else:
                raise StopIteration()

        is_last_row = False
        new_index = index + 1
        if new_index < len(data_batch):
            # Update the index in the dictionary.
            self.list_timer_iterator_cursors[iterator_id] = (
                data_batch,
                new_index,
                require_next_fetch,
            )
        else:
            # If the index is at the end of the data batch, remove the state from the dictionary.
            self.list_timer_iterator_cursors.pop(iterator_id, None)
            is_last_row = True

        is_last_row_from_iterator = is_last_row and not require_next_fetch
        timestamp = data_batch[index]
        return (timestamp, is_last_row_from_iterator)

    def get_expiry_timers_iterator(
        self, iterator_id: str, expiry_timestamp: int
    ) -> Tuple[Tuple, int, bool]:
        import pyspark.sql.streaming.proto.StateMessage_pb2 as stateMessage

        if iterator_id in self.expiry_timer_iterator_cursors:
            # If the state is already in the dictionary, return the next row.
            data_batch, index, require_next_fetch = self.expiry_timer_iterator_cursors[iterator_id]
        else:
            expiry_timer_call = stateMessage.ExpiryTimerRequest(
                expiryTimestampMs=expiry_timestamp, iteratorId=iterator_id
            )
            timer_request = stateMessage.TimerRequest(expiryTimerRequest=expiry_timer_call)
            message = stateMessage.StateRequest(timerRequest=timer_request)

            self._send_proto_message(message.SerializeToString())
            response_message = self._receive_proto_message_with_timers()
            status = response_message[0]
            if status == 0:
                data_batch = list(
                    map(
                        lambda x: (self._deserialize_from_bytes(x.key), x.timestampMs),
                        response_message[2],
                    )
                )
                require_next_fetch = response_message[3]
                index = 0
            else:
                raise StopIteration()

        is_last_row = False
        new_index = index + 1
        if new_index < len(data_batch):
            # Update the index in the dictionary.
            self.expiry_timer_iterator_cursors[iterator_id] = (
                data_batch,
                new_index,
                require_next_fetch,
            )
        else:
            # If the index is at the end of the data batch, remove the state from the dictionary.
            self.expiry_timer_iterator_cursors.pop(iterator_id, None)
            is_last_row = True

        is_last_row_from_iterator = is_last_row and not require_next_fetch
        key, timestamp = data_batch[index]
        return (key, timestamp, is_last_row_from_iterator)

    def get_timestamps(self, time_mode: str) -> Tuple[int, int]:
        if time_mode.lower() == "none":
            return -1, -1
        else:
            if self._batch_timestamp == -1:
                self._batch_timestamp = self._get_batch_timestamp()
            if self._watermark_timestamp == -1:
                self._watermark_timestamp = self._get_watermark_timestamp()
        return self._batch_timestamp, self._watermark_timestamp

    def get_map_state(
        self,
        state_name: str,
        user_key_schema: Union[StructType, str],
        value_schema: Union[StructType, str],
        ttl_duration_ms: Optional[int],
    ) -> None:
        import pyspark.sql.streaming.proto.StateMessage_pb2 as stateMessage

        if isinstance(user_key_schema, str):
            user_key_schema = self._parse_string_schema(user_key_schema)
        if isinstance(value_schema, str):
            value_schema = self._parse_string_schema(value_schema)

        state_call_command = stateMessage.StateCallCommand()
        state_call_command.stateName = state_name
        state_call_command.schema = user_key_schema.json()
        state_call_command.mapStateValueSchema = value_schema.json()
        if ttl_duration_ms is not None:
            state_call_command.ttl.durationMs = ttl_duration_ms
        call = stateMessage.StatefulProcessorCall(getMapState=state_call_command)
        message = stateMessage.StateRequest(statefulProcessorCall=call)

        self._send_proto_message(message.SerializeToString())
        response_message = self._receive_proto_message()
        status = response_message[0]
        if status != 0:
            # TODO(SPARK-49233): Classify user facing errors.
            raise PySparkRuntimeError(f"Error initializing map state: {response_message[1]}")

    def delete_if_exists(self, state_name: str) -> None:
        import pyspark.sql.streaming.proto.StateMessage_pb2 as stateMessage

        state_call_command = stateMessage.StateCallCommand()
        state_call_command.stateName = state_name
        call = stateMessage.StatefulProcessorCall(deleteIfExists=state_call_command)
        message = stateMessage.StateRequest(statefulProcessorCall=call)

        self._send_proto_message(message.SerializeToString())
        response_message = self._receive_proto_message()
        status = response_message[0]
        if status != 0:
            # TODO(SPARK-49233): Classify user facing errors.
            raise PySparkRuntimeError(f"Error deleting state: {response_message[1]}")

    def _get_batch_timestamp(self) -> int:
        import pyspark.sql.streaming.proto.StateMessage_pb2 as stateMessage

        get_processing_time_call = stateMessage.GetProcessingTime()
        timer_value_call = stateMessage.TimerValueRequest(
            getProcessingTimer=get_processing_time_call
        )
        timer_request = stateMessage.TimerRequest(timerValueRequest=timer_value_call)
        message = stateMessage.StateRequest(timerRequest=timer_request)

        self._send_proto_message(message.SerializeToString())
        response_message = self._receive_proto_message_with_long_value()
        status = response_message[0]
        if status != 0:
            # TODO(SPARK-49233): Classify user facing errors.
            raise PySparkRuntimeError(f"Error getting processing timestamp: {response_message[1]}")
        else:
            timestamp = response_message[2]
            return timestamp

    def _get_watermark_timestamp(self) -> int:
        import pyspark.sql.streaming.proto.StateMessage_pb2 as stateMessage

        get_watermark_call = stateMessage.GetWatermark()
        timer_value_call = stateMessage.TimerValueRequest(getWatermark=get_watermark_call)
        timer_request = stateMessage.TimerRequest(timerValueRequest=timer_value_call)
        message = stateMessage.StateRequest(timerRequest=timer_request)

        self._send_proto_message(message.SerializeToString())
        response_message = self._receive_proto_message_with_long_value()
        status = response_message[0]
        if status != 0:
            # TODO(SPARK-49233): Classify user facing errors.
            raise PySparkRuntimeError(f"Error getting eventtime timestamp: {response_message[1]}")
        else:
            timestamp = response_message[2]
            return timestamp

    def _send_proto_message(self, message: bytes) -> None:
        # Writing zero here to indicate message version. This allows us to evolve the message
        # format or even changing the message protocol in the future.
        write_int(0, self.sockfile)
        write_int(len(message), self.sockfile)
        self.sockfile.write(message)
        self.sockfile.flush()

    def _receive_proto_message(self) -> Tuple[int, str, bytes]:
        import pyspark.sql.streaming.proto.StateMessage_pb2 as stateMessage

        length = read_int(self.sockfile)
        bytes = self.sockfile.read(length)
        message = stateMessage.StateResponse()
        message.ParseFromString(bytes)
        return message.statusCode, message.errorMessage, message.value

    def _receive_proto_message_with_long_value(self) -> Tuple[int, str, int]:
        import pyspark.sql.streaming.proto.StateMessage_pb2 as stateMessage

        length = read_int(self.sockfile)
        bytes = self.sockfile.read(length)
        message = stateMessage.StateResponseWithLongTypeVal()
        message.ParseFromString(bytes)
        return message.statusCode, message.errorMessage, message.value

    def _receive_proto_message_with_string_value(self) -> Tuple[int, str, str]:
        import pyspark.sql.streaming.proto.StateMessage_pb2 as stateMessage

        length = read_int(self.sockfile)
        bytes = self.sockfile.read(length)
        message = stateMessage.StateResponseWithStringTypeVal()
        message.ParseFromString(bytes)
        return message.statusCode, message.errorMessage, message.value

    # The third return type is RepeatedScalarFieldContainer[bytes], which is protobuf's container
    # type. We simplify it to Any here to avoid unnecessary complexity.
    def _receive_proto_message_with_list_get(self) -> Tuple[int, str, Any, bool]:
        import pyspark.sql.streaming.proto.StateMessage_pb2 as stateMessage

        length = read_int(self.sockfile)
        bytes = self.sockfile.read(length)
        message = stateMessage.StateResponseWithListGet()
        message.ParseFromString(bytes)

        return message.statusCode, message.errorMessage, message.value, message.requireNextFetch

    # The third return type is RepeatedScalarFieldContainer[bytes], which is protobuf's container
    # type. We simplify it to Any here to avoid unnecessary complexity.
    def _receive_proto_message_with_map_keys_values(self) -> Tuple[int, str, Any, bool]:
        import pyspark.sql.streaming.proto.StateMessage_pb2 as stateMessage

        length = read_int(self.sockfile)
        bytes = self.sockfile.read(length)
        message = stateMessage.StateResponseWithMapKeysOrValues()
        message.ParseFromString(bytes)

        return message.statusCode, message.errorMessage, message.value, message.requireNextFetch

    # The third return type is RepeatedScalarFieldContainer[KeyAndValuePair], which is protobuf's
    # container type. We simplify it to Any here to avoid unnecessary complexity.
    def _receive_proto_message_with_map_pairs(self) -> Tuple[int, str, Any, bool]:
        import pyspark.sql.streaming.proto.StateMessage_pb2 as stateMessage

        length = read_int(self.sockfile)
        bytes = self.sockfile.read(length)
        message = stateMessage.StateResponseWithMapIterator()
        message.ParseFromString(bytes)

        return message.statusCode, message.errorMessage, message.kvPair, message.requireNextFetch

    # The third return type is RepeatedScalarFieldContainer[TimerInfo], which is protobuf's
    # container type. We simplify it to Any here to avoid unnecessary complexity.
    def _receive_proto_message_with_timers(self) -> Tuple[int, str, Any, bool]:
        import pyspark.sql.streaming.proto.StateMessage_pb2 as stateMessage

        length = read_int(self.sockfile)
        bytes = self.sockfile.read(length)
        message = stateMessage.StateResponseWithTimer()
        message.ParseFromString(bytes)

        return message.statusCode, message.errorMessage, message.timer, message.requireNextFetch

    def _receive_str(self) -> str:
        return self.utf8_deserializer.loads(self.sockfile)

    def _serialize_to_bytes(self, schema: StructType, data: Tuple) -> bytes:
        from pyspark.testing.utils import have_numpy

        if have_numpy:
            import numpy as np

            def normalize_value(v: Any) -> Any:
                # Convert NumPy types to Python primitive types.
                if isinstance(v, np.generic):
                    return v.tolist()
                # Named tuples (collections.namedtuple or typing.NamedTuple) and Row both
                # require positional arguments and cannot be instantiated
                # with a generator expression.
                if isinstance(v, Row) or (isinstance(v, tuple) and hasattr(v, "_fields")):
                    return type(v)(*[normalize_value(e) for e in v])
                # List / tuple: recursively normalize each element
                if isinstance(v, (list, tuple)):
                    return type(v)(normalize_value(e) for e in v)
                # Dict: normalize both keys and values
                if isinstance(v, dict):
                    return {normalize_value(k): normalize_value(val) for k, val in v.items()}
                # Address a couple of pandas dtypes too.
                elif hasattr(v, "to_pytimedelta"):
                    return v.to_pytimedelta()
                elif hasattr(v, "to_pydatetime"):
                    return v.to_pydatetime()
                else:
                    return v

            converted = tuple(normalize_value(v) for v in data)
        else:
            converted = data

        return self.pickleSer.dumps(schema.toInternal(converted))

    def _deserialize_from_bytes(self, value: bytes) -> Any:
        return self.pickleSer.loads(value)

    def _send_arrow_state(self, schema: StructType, state: List[Tuple]) -> None:
        import pyarrow as pa
        import pandas as pd

        column_names = [field.name for field in schema.fields]
        pandas_df = convert_pandas_using_numpy_type(
            pd.DataFrame(state, columns=column_names), schema
        )
        batch = pa.RecordBatch.from_pandas(pandas_df)
        self.serializer.dump_stream([batch], cast(IO[bytes], self.sockfile))
        self.sockfile.flush()

    def _read_arrow_state(self) -> Any:
        return self.serializer.load_stream(cast(IO[bytes], self.sockfile))

    def _send_list_state(self, schema: StructType, state: List[Tuple]) -> None:
        for value in state:
            bytes = self._serialize_to_bytes(schema, value)
            length = len(bytes)
            write_int(length, self.sockfile)
            self.sockfile.write(bytes)

        write_int(-1, self.sockfile)
        self.sockfile.flush()

    def _read_list_state(self) -> List[Any]:
        data_array = []
        while True:
            length = read_int(self.sockfile)
            if length < 0:
                break
            bytes = self.sockfile.read(length)
            data_array.append(self._deserialize_from_bytes(bytes))
        return data_array

    # Parse a string schema into a StructType schema. This method will perform an API call to
    # JVM side to parse the schema string.
    def _parse_string_schema(self, schema: str) -> StructType:
        import pyspark.sql.streaming.proto.StateMessage_pb2 as stateMessage

        parse_string_schema_call = stateMessage.ParseStringSchema(schema=schema)
        utils_request = stateMessage.UtilsRequest(parseStringSchema=parse_string_schema_call)
        message = stateMessage.StateRequest(utilsRequest=utils_request)

        self._send_proto_message(message.SerializeToString())
        response_message = self._receive_proto_message_with_string_value()
        status = response_message[0]
        if status != 0:
            # TODO(SPARK-49233): Classify user facing errors.
            raise PySparkRuntimeError(f"Error parsing string schema: {response_message[1]}")
        else:
            return StructType.fromJson(json.loads(response_message[2]))


class ListTimerIterator:
    def __init__(self, stateful_processor_api_client: StatefulProcessorApiClient):
        # Generate a unique identifier for the iterator to make sure iterators on the
        # same partition won't interfere with each other
        self.iterator_id = str(uuid.uuid4())
        self.stateful_processor_api_client = stateful_processor_api_client
        self.iterator_fully_consumed = False

    def __iter__(self) -> Iterator[int]:
        return self

    def __next__(self) -> int:
        if self.iterator_fully_consumed:
            raise StopIteration()

        ts, is_last_row = self.stateful_processor_api_client.get_list_timer_row(self.iterator_id)
        if is_last_row:
            self.iterator_fully_consumed = True

        return ts


class ExpiredTimerIterator:
    def __init__(
        self, stateful_processor_api_client: StatefulProcessorApiClient, expiry_timestamp: int
    ):
        # Generate a unique identifier for the iterator to make sure iterators on the
        # same partition won't interfere with each other
        self.iterator_id = str(uuid.uuid4())
        self.stateful_processor_api_client = stateful_processor_api_client
        self.expiry_timestamp = expiry_timestamp
        self.iterator_fully_consumed = False

    def __iter__(self) -> Iterator[Tuple[Tuple, int]]:
        return self

    def __next__(self) -> Tuple[Tuple, int]:
        if self.iterator_fully_consumed:
            raise StopIteration()

        key, ts, is_last_row = self.stateful_processor_api_client.get_expiry_timers_iterator(
            self.iterator_id, self.expiry_timestamp
        )
        if is_last_row:
            self.iterator_fully_consumed = True

        return (key, ts)



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import gc
import os
import sys
from tempfile import NamedTemporaryFile
import threading
import pickle
from typing import (
    overload,
    Any,
    BinaryIO,
    Callable,
    Dict,
    Generic,
    IO,
    Iterator,
    Optional,
    Tuple,
    TypeVar,
    TYPE_CHECKING,
    Union,
)

from pyspark.serializers import ChunkedStream, pickle_protocol
from pyspark.util import print_exec, local_connect_and_auth
from pyspark.errors import PySparkRuntimeError

if TYPE_CHECKING:
    from pyspark import SparkContext


__all__ = ["Broadcast"]

T = TypeVar("T")


# Holds broadcasted data received from Java, keyed by its id.
_broadcastRegistry: Dict[int, "Broadcast[Any]"] = {}


def _from_id(bid: int) -> "Broadcast[Any]":
    from pyspark.core.broadcast import _broadcastRegistry

    if bid not in _broadcastRegistry:
        raise PySparkRuntimeError(
            errorClass="BROADCAST_VARIABLE_NOT_LOADED",
            messageParameters={
                "variable": str(bid),
            },
        )
    return _broadcastRegistry[bid]


class Broadcast(Generic[T]):
    """
    A broadcast variable created with :meth:`SparkContext.broadcast`.
    Access its value through :attr:`value`.

    Examples
    --------
    >>> b = spark.sparkContext.broadcast([1, 2, 3, 4, 5])
    >>> b.value
    [1, 2, 3, 4, 5]
    >>> spark.sparkContext.parallelize([0, 0]).flatMap(lambda x: b.value).collect()
    [1, 2, 3, 4, 5, 1, 2, 3, 4, 5]
    >>> b.unpersist()

    >>> large_broadcast = spark.sparkContext.broadcast(range(10000))
    """

    @overload  # On driver
    def __init__(
        self: "Broadcast[T]",
        sc: "SparkContext",
        value: T,
        pickle_registry: "BroadcastPickleRegistry",
    ): ...

    @overload  # On worker without decryption server
    def __init__(self: "Broadcast[Any]", *, path: str): ...

    @overload  # On worker with decryption server
    def __init__(self: "Broadcast[Any]", *, sock_file: str): ...

    def __init__(  # type: ignore[misc]
        self,
        sc: Optional["SparkContext"] = None,
        value: Optional[T] = None,
        pickle_registry: Optional["BroadcastPickleRegistry"] = None,
        path: Optional[str] = None,
        sock_file: Optional[BinaryIO] = None,
    ):
        """
        Should not be called directly by users -- use :meth:`SparkContext.broadcast`
        instead.
        """
        if sc is not None:
            # we're on the driver.  We want the pickled data to end up in a file (maybe encrypted)
            f = NamedTemporaryFile(delete=False, dir=sc._temp_dir)
            self._path = f.name
            self._sc: Optional["SparkContext"] = sc
            assert sc._jvm is not None
            self._python_broadcast = sc._jvm.PythonRDD.setupBroadcast(self._path)
            broadcast_out: Union[ChunkedStream, IO[bytes]]
            if sc._encryption_enabled:
                # with encryption, we ask the jvm to do the encryption for us, we send it data
                # over a socket
                conn_info, auth_secret = self._python_broadcast.setupEncryptionServer()
                encryption_sock_file, _ = local_connect_and_auth(conn_info, auth_secret)
                broadcast_out = ChunkedStream(encryption_sock_file, 8192)
            else:
                # no encryption, we can just write pickled data directly to the file from python
                broadcast_out = f
            self.dump(value, broadcast_out)  # type: ignore[arg-type]
            if sc._encryption_enabled:
                self._python_broadcast.waitTillDataReceived()
            self._jbroadcast = sc._jsc.broadcast(self._python_broadcast)
            self._pickle_registry = pickle_registry
        else:
            # we're on an executor
            self._jbroadcast = None
            self._sc = None
            self._python_broadcast = None
            if sock_file is not None:
                # the jvm is doing decryption for us.  Read the value
                # immediately from the sock_file
                self._value = self.load(sock_file)
            else:
                # the jvm just dumps the pickled data in path -- we'll unpickle lazily when
                # the value is requested
                assert path is not None
                self._path = path

    def dump(self, value: T, f: BinaryIO) -> None:
        """
        Write a pickled representation of value to the open file or socket.
        The protocol pickle is HIGHEST_PROTOCOL.

        Parameters
        ----------
        value : T
            Value to write.

        f : :class:`BinaryIO`
            File or socket where the pickled value will be stored.

        Examples
        --------
        >>> import os
        >>> import tempfile

        >>> b = spark.sparkContext.broadcast([1, 2, 3, 4, 5])

        Write a pickled representation of `b` to the open temp file.

        >>> with tempfile.TemporaryDirectory(prefix="dump") as d:
        ...     path = os.path.join(d, "test.txt")
        ...     with open(path, "wb") as f:
        ...         b.dump(b.value, f)
        """
        try:
            pickle.dump(value, f, pickle_protocol)
        except pickle.PickleError:
            raise
        except Exception as e:
            msg = "Could not serialize broadcast: %s: %s" % (e.__class__.__name__, str(e))
            print_exec(sys.stderr)
            raise pickle.PicklingError(msg)
        f.close()

    def load_from_path(self, path: str) -> T:
        """
        Read the pickled representation of an object from the open file and
        return the reconstituted object hierarchy specified therein.

        Parameters
        ----------
        path : str
            File path where reads the pickled value.

        Returns
        -------
        T
            The object hierarchy specified therein reconstituted
            from the pickled representation of an object.

        Examples
        --------
        >>> import os
        >>> import tempfile

        >>> b = spark.sparkContext.broadcast([1, 2, 3, 4, 5])
        >>> c = spark.sparkContext.broadcast(1)

        Read the pickled representation of value from temp file.

        >>> with tempfile.TemporaryDirectory(prefix="load_from_path") as d:
        ...     path = os.path.join(d, "test.txt")
        ...     with open(path, "wb") as f:
        ...         b.dump(b.value, f)
        ...     c.load_from_path(path)
        [1, 2, 3, 4, 5]
        """
        with open(path, "rb", 1 << 20) as f:
            return self.load(f)

    def load(self, file: BinaryIO) -> T:
        """
        Read a pickled representation of value from the open file or socket.

        Parameters
        ----------
        file : :class:`BinaryIO`
            File or socket where the pickled value will be read.

        Returns
        -------
        T
            The object hierarchy specified therein reconstituted
            from the pickled representation of an object.

        Examples
        --------
        >>> import os
        >>> import tempfile

        >>> b = spark.sparkContext.broadcast([1, 2, 3, 4, 5])
        >>> c = spark.sparkContext.broadcast(1)

        Read the pickled representation of value from the open temp file.

        >>> with tempfile.TemporaryDirectory(prefix="load") as d:
        ...     path = os.path.join(d, "test.txt")
        ...     with open(path, "wb") as f:
        ...         b.dump(b.value, f)
        ...     with open(path, "rb") as f:
        ...         c.load(f)
        [1, 2, 3, 4, 5]
        """
        gc.disable()
        try:
            return pickle.load(file)
        finally:
            gc.enable()

    @property
    def value(self) -> T:
        """Return the broadcasted value"""
        if not hasattr(self, "_value") and self._path is not None:
            # we only need to decrypt it here when encryption is enabled and
            # if its on the driver, since executor decryption is handled already
            if self._sc is not None and self._sc._encryption_enabled:
                conn_info, auth_secret = self._python_broadcast.setupDecryptionServer()
                decrypted_sock_file, _ = local_connect_and_auth(conn_info, auth_secret)
                self._python_broadcast.waitTillBroadcastDataSent()
                return self.load(decrypted_sock_file)
            else:
                self._value = self.load_from_path(self._path)
        return self._value

    def unpersist(self, blocking: bool = False) -> None:
        """
        Delete cached copies of this broadcast on the executors. If the
        broadcast is used after this is called, it will need to be
        re-sent to each executor.

        Parameters
        ----------
        blocking : bool, optional, default False
            Whether to block until unpersisting has completed.

        Examples
        --------
        >>> b = spark.sparkContext.broadcast([1, 2, 3, 4, 5])

        Delete cached copies of this broadcast on the executors

        >>> b.unpersist()
        """
        if self._jbroadcast is None:
            raise PySparkRuntimeError(
                errorClass="INVALID_BROADCAST_OPERATION",
                messageParameters={"operation": "unpersisted"},
            )
        self._jbroadcast.unpersist(blocking)

    def destroy(self, blocking: bool = False) -> None:
        """
        Destroy all data and metadata related to this broadcast variable.
        Use this with caution; once a broadcast variable has been destroyed,
        it cannot be used again.

        .. versionchanged:: 3.0.0
           Added optional argument `blocking` to specify whether to block until all
           blocks are deleted.

        Parameters
        ----------
        blocking : bool, optional, default False
            Whether to block until unpersisting has completed.

        Examples
        --------
        >>> b = spark.sparkContext.broadcast([1, 2, 3, 4, 5])

        Destroy all data and metadata related to this broadcast variable

        >>> b.destroy()
        """
        if self._jbroadcast is None:
            raise PySparkRuntimeError(
                errorClass="INVALID_BROADCAST_OPERATION",
                messageParameters={"operation": "destroyed"},
            )
        self._jbroadcast.destroy(blocking)
        os.unlink(self._path)

    def __reduce__(self) -> Tuple[Callable[[int], "Broadcast[T]"], Tuple[int]]:
        if self._jbroadcast is None:
            raise PySparkRuntimeError(
                errorClass="INVALID_BROADCAST_OPERATION",
                messageParameters={"operation": "serialized"},
            )
        assert self._pickle_registry is not None
        self._pickle_registry.add(self)
        return _from_id, (self._jbroadcast.id(),)


class BroadcastPickleRegistry(threading.local):
    """Thread-local registry for broadcast variables that have been pickled"""

    def __init__(self) -> None:
        self.__dict__.setdefault("_registry", set())

    def __iter__(self) -> Iterator[Broadcast[Any]]:
        for bcast in self._registry:
            yield bcast

    def add(self, bcast: Broadcast[Any]) -> None:
        self._registry.add(bcast)

    def clear(self) -> None:
        self._registry.clear()


def _test() -> None:
    import doctest
    from pyspark.sql import SparkSession
    import pyspark.core.broadcast

    globs = pyspark.core.broadcast.__dict__.copy()
    spark = SparkSession.builder.master("local[4]").appName("broadcast tests").getOrCreate()
    globs["spark"] = spark

    failure_count, test_count = doctest.testmod(pyspark.core.broadcast, globs=globs)
    spark.stop()
    if failure_count:
        sys.exit(-1)


if __name__ == "__main__":
    _test()



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
import pandas as pd

from pyspark import pandas as ps
from pyspark.loose_version import LooseVersion
from pyspark.pandas.exceptions import SparkPandasIndexingError, SparkPandasNotImplementedError
from pyspark.testing.pandasutils import PandasOnSparkTestCase
from pyspark.testing.sqlutils import SQLTestUtils


class IndexingILocMixin:
    @property
    def pdf(self):
        return pd.DataFrame(
            {"a": [1, 2, 3, 4, 5, 6, 7, 8, 9], "b": [4, 5, 6, 3, 2, 1, 0, 0, 0]},
            index=[0, 1, 3, 5, 6, 8, 9, 9, 9],
        )

    @property
    def pdf2(self):
        return pd.DataFrame(
            {0: [1, 2, 3, 4, 5, 6, 7, 8, 9], 1: [4, 5, 6, 3, 2, 1, 0, 0, 0]},
            index=[0, 1, 3, 5, 6, 8, 9, 9, 9],
        )

    @property
    def psdf(self):
        return ps.from_pandas(self.pdf)

    @property
    def psdf2(self):
        return ps.from_pandas(self.pdf2)

    def test_iloc(self):
        pdf = pd.DataFrame({"A": [1, 2], "B": [3, 4], "C": [5, 6]})
        psdf = ps.from_pandas(pdf)

        self.assert_eq(psdf.iloc[0, 0], pdf.iloc[0, 0])
        for indexer in [0, [0], [0, 1], [1, 0], [False, True, True], slice(0, 1)]:
            self.assert_eq(psdf.iloc[:, indexer], pdf.iloc[:, indexer])
            self.assert_eq(psdf.iloc[:1, indexer], pdf.iloc[:1, indexer])
            self.assert_eq(psdf.iloc[:-1, indexer], pdf.iloc[:-1, indexer])
            # self.assert_eq(psdf.iloc[psdf.index == 2, indexer], pdf.iloc[pdf.index == 2, indexer])

        self.assertRaisesRegex(
            SparkPandasNotImplementedError,
            ".iloc requires numeric slice, conditional boolean",
            lambda: ps.range(10).iloc["a", :],
        )

    def test_iloc_multiindex_columns(self):
        arrays = [np.array(["bar", "bar", "baz", "baz"]), np.array(["one", "two", "one", "two"])]

        pdf = pd.DataFrame(np.random.randn(3, 4), index=["A", "B", "C"], columns=arrays)
        psdf = ps.from_pandas(pdf)

        for indexer in [0, [0], [0, 1], [1, 0], [False, True, True, True], slice(0, 1)]:
            self.assert_eq(psdf.iloc[:, indexer], pdf.iloc[:, indexer])
            self.assert_eq(psdf.iloc[:1, indexer], pdf.iloc[:1, indexer])
            self.assert_eq(psdf.iloc[:-1, indexer], pdf.iloc[:-1, indexer])
            # self.assert_eq(psdf.iloc[psdf.index == "B", indexer],
            #                pdf.iloc[pdf.index == "B", indexer])

    def test_iloc_series(self):
        pser = pd.Series([1, 2, 3])
        psser = ps.from_pandas(pser)

        self.assert_eq(psser.iloc[0], pser.iloc[0])
        self.assert_eq(psser.iloc[:], pser.iloc[:])
        self.assert_eq(psser.iloc[:1], pser.iloc[:1])
        self.assert_eq(psser.iloc[:-1], pser.iloc[:-1])

        self.assert_eq((psser + 1).iloc[0], (pser + 1).iloc[0])
        self.assert_eq((psser + 1).iloc[:], (pser + 1).iloc[:])
        self.assert_eq((psser + 1).iloc[:1], (pser + 1).iloc[:1])
        self.assert_eq((psser + 1).iloc[:-1], (pser + 1).iloc[:-1])

    def test_iloc_slice_rows_sel(self):
        pdf = pd.DataFrame({"A": [1, 2] * 5, "B": [3, 4] * 5, "C": [5, 6] * 5})
        psdf = ps.from_pandas(pdf)

        for rows_sel in [
            slice(None),
            slice(0, 1),
            slice(1, 2),
            slice(-3, None),
            slice(None, -3),
            slice(None, 0),
            slice(None, None, 3),
            slice(3, 8, 2),
            slice(None, None, -2),
            slice(8, 3, -2),
            slice(8, None, -2),
            slice(None, 3, -2),
        ]:
            with self.subTest(rows_sel=rows_sel):
                self.assert_eq(psdf.iloc[rows_sel].sort_index(), pdf.iloc[rows_sel].sort_index())
                self.assert_eq(
                    psdf.A.iloc[rows_sel].sort_index(), pdf.A.iloc[rows_sel].sort_index()
                )
                self.assert_eq(
                    (psdf.A + 1).iloc[rows_sel].sort_index(),
                    (pdf.A + 1).iloc[rows_sel].sort_index(),
                )

    def test_iloc_iterable_rows_sel(self):
        pdf = pd.DataFrame({"A": [1, 2] * 5, "B": [3, 4] * 5, "C": [5, 6] * 5})
        psdf = ps.from_pandas(pdf)

        for rows_sel in [
            [],
            np.array([0, 1]),
            [1, 2],
            np.array([-3]),
            [3],
            np.array([-2]),
            [8, 3, -5],
        ]:
            with self.subTest(rows_sel=rows_sel):
                self.assert_eq(psdf.iloc[rows_sel].sort_index(), pdf.iloc[rows_sel].sort_index())
                self.assert_eq(
                    psdf.A.iloc[rows_sel].sort_index(), pdf.A.iloc[rows_sel].sort_index()
                )
                self.assert_eq(
                    (psdf.A + 1).iloc[rows_sel].sort_index(),
                    (pdf.A + 1).iloc[rows_sel].sort_index(),
                )

            with self.subTest(rows_sel=rows_sel):
                self.assert_eq(
                    psdf.iloc[rows_sel, :].sort_index(), pdf.iloc[rows_sel, :].sort_index()
                )

            with self.subTest(rows_sel=rows_sel):
                self.assert_eq(
                    psdf.iloc[rows_sel, :1].sort_index(), pdf.iloc[rows_sel, :1].sort_index()
                )

    def test_frame_iloc_setitem(self):
        pdf = pd.DataFrame(
            [[1, 2], [4, 5], [7, 8]],
            index=["cobra", "viper", "sidewinder"],
            columns=["max_speed", "shield"],
        )
        psdf = ps.from_pandas(pdf)

        pdf.iloc[[1, 2], [1, 0]] = 10
        psdf.iloc[[1, 2], [1, 0]] = 10
        self.assert_eq(psdf, pdf)

        pdf.iloc[0, 1] = 50
        psdf.iloc[0, 1] = 50
        self.assert_eq(psdf, pdf)

        with self.assertRaisesRegex(ValueError, "setting an array element with a sequence."):
            psdf.iloc[0, 0] = -psdf.max_speed
        with self.assertRaisesRegex(ValueError, "shape mismatch"):
            psdf.iloc[:, [1, 0]] = -psdf.max_speed
        with self.assertRaisesRegex(ValueError, "Only a dataframe with one column can be assigned"):
            psdf.iloc[:, 0] = psdf

        pdf = pd.DataFrame(
            [[1], [4], [7]], index=["cobra", "viper", "sidewinder"], columns=["max_speed"]
        )
        psdf = ps.from_pandas(pdf)

        if LooseVersion(pd.__version__) < "3.0.0":
            pdf.iloc[:, 0] = pdf
            psdf.iloc[:, 0] = psdf
            self.assert_eq(psdf, pdf)

    def test_series_iloc_setitem(self):
        pdf = pd.DataFrame({"x": [1, 2, 3], "y": [4, 5, 6]}, index=["cobra", "viper", "sidewinder"])
        psdf = ps.from_pandas(pdf)

        pser = pdf.x
        psery = pdf.y
        psser = psdf.x
        pssery = psdf.y

        piloc = pser.iloc
        kiloc = psser.iloc

        pser1 = pser + 1
        psser1 = psser + 1

        for key, value in [
            ([1, 2], 10),
            (1, 50),
            (slice(None), 10),
            (slice(None, 1), 20),
            (slice(1, None), 30),
        ]:
            with self.subTest(key=key, value=value):
                pser.iloc[key] = value
                psser.iloc[key] = value
                self.assert_eq(psser, pser)
                self.assert_eq(psdf, pdf)
                self.assert_eq(pssery, psery)

                piloc[key] = -value
                kiloc[key] = -value
                self.assert_eq(psser, pser)
                self.assert_eq(psdf, pdf)
                self.assert_eq(pssery, psery)

                pser1.iloc[key] = value
                psser1.iloc[key] = value
                self.assert_eq(psser1, pser1)
                self.assert_eq(psdf, pdf)
                self.assert_eq(pssery, psery)

        with self.assertRaises(ValueError):
            psser.iloc[1] = -psser

        pser = pd.Index([1, 2, 3]).to_series()
        psser = ps.Index([1, 2, 3]).to_series()

        pser1 = pser + 1
        psser1 = psser + 1

        pser.iloc[0] = 10
        psser.iloc[0] = 10
        self.assert_eq(psser, pser)

        pser1.iloc[0] = 20
        psser1.iloc[0] = 20
        self.assert_eq(psser1, pser1)

        pdf = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
        psdf = ps.from_pandas(pdf)

        pser = pdf.a
        psser = psdf.a

        pser.iloc[[0, 1, 2]] = -pdf.b
        psser.iloc[[0, 1, 2]] = -psdf.b
        self.assert_eq(psser, pser)
        self.assert_eq(psdf, pdf)

        with self.assertRaisesRegex(ValueError, "setting an array element with a sequence."):
            psser.iloc[1] = psdf[["b"]]

    def test_iloc_raises(self):
        pdf = pd.DataFrame({"A": [1, 2], "B": [3, 4], "C": [5, 6]})
        psdf = ps.from_pandas(pdf)

        with self.assertRaisesRegex(SparkPandasIndexingError, "Only accepts pairs of candidates"):
            psdf.iloc[[0, 1], [0, 1], [1, 2]]

        with self.assertRaisesRegex(SparkPandasIndexingError, "Too many indexers"):
            psdf.A.iloc[[0, 1], [0, 1]]

        with self.assertRaisesRegex(TypeError, "cannot do slice indexing with these indexers"):
            psdf.iloc[:"b", :]

        with self.assertRaisesRegex(TypeError, "cannot do slice indexing with these indexers"):
            psdf.iloc[:, :"b"]

        with self.assertRaisesRegex(TypeError, "cannot perform reduce with flexible type"):
            psdf.iloc[:, ["A"]]

        with self.assertRaisesRegex(ValueError, "Location based indexing can only have"):
            psdf.iloc[:, "A"]

        with self.assertRaisesRegex(IndexError, "out of range"):
            psdf.iloc[:, [5, 6]]


class IndexingILocTests(
    IndexingILocMixin,
    PandasOnSparkTestCase,
    SQLTestUtils,
):
    pass


if __name__ == "__main__":
    from pyspark.testing import main

    main()



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import random
import tempfile
import unittest

import pandas as pd
import pandas.testing as pdt

from pyspark import SparkConf
from pyspark.sql import DataFrame
from pyspark.sql.functions import split
from pyspark.sql.streaming import StatefulProcessor, TwsTester
from pyspark.sql.streaming.query import StreamingQuery
from pyspark.sql.tests.pandas.helper.helper_pandas_transform_with_state import (
    AllMethodsTestProcessorFactory,
    EventTimeCountProcessorFactory,
    EventTimeSessionProcessorFactory,
    RunningCountStatefulProcessorFactory,
    SessionTimeoutProcessorFactory,
    TopKProcessorFactory,
    WordFrequencyProcessorFactory,
)
from pyspark.sql.types import (
    DoubleType,
    IntegerType,
    Row,
    StringType,
    StructField,
    StructType,
)
from pyspark.errors import PySparkValueError, PySparkAssertionError
from pyspark.errors.exceptions.base import IllegalArgumentException
from pyspark.testing.sqlutils import ReusedSQLTestCase
from pyspark.testing.utils import (
    have_pandas,
    have_pyarrow,
    pandas_requirement_message,
    pyarrow_requirement_message,
)


@unittest.skipIf(
    not have_pandas or not have_pyarrow or os.environ.get("PYTHON_GIL", "?") == "0",
    pandas_requirement_message or pyarrow_requirement_message or "Not supported in no-GIL mode",
)
class TwsTesterTests(ReusedSQLTestCase):
    def test_running_count_processor(self):
        processor = RunningCountStatefulProcessorFactory().row()
        tester = TwsTester(processor)

        self.assertEqual(tester.test("key1", [Row(value="a")]), [Row(key="key1", count=1)])
        self.assertEqual(
            tester.test("key2", [Row(value="a"), Row(value="a")]),
            [Row(key="key2", count=2)],
        )
        self.assertEqual(tester.test("key3", [Row(value="a")]), [Row(key="key3", count=1)])
        self.assertEqual(
            tester.test("key1", [Row(value="a"), Row(value="a"), Row(value="a")]),
            [Row(key="key1", count=4)],
        )

        self.assertEqual(tester.peekValueState("count", "key1"), (4,))
        self.assertEqual(tester.peekValueState("count", "key2"), (2,))
        self.assertEqual(tester.peekValueState("count", "key3"), (1,))
        self.assertIsNone(tester.peekValueState("count", "key4"))

    def test_running_count_processor_pandas(self):
        processor = RunningCountStatefulProcessorFactory().pandas()
        tester = TwsTester(processor)

        ans1 = tester.testInPandas("key1", pd.DataFrame({"value": ["a"]}))
        expected1 = pd.DataFrame({"key": ["key1"], "count": [1]})
        pdt.assert_frame_equal(ans1, expected1, check_like=True)

        ans2 = tester.testInPandas("key2", pd.DataFrame({"value": ["a", "a"]}))
        expected2 = pd.DataFrame({"key": ["key2"], "count": [2]})
        pdt.assert_frame_equal(ans2, expected2, check_like=True)

        ans3 = tester.testInPandas("key3", pd.DataFrame({"value": ["a"]}))
        expected3 = pd.DataFrame({"key": ["key3"], "count": [1]})
        pdt.assert_frame_equal(ans3, expected3, check_like=True)

        ans4 = tester.testInPandas("key1", pd.DataFrame({"value": ["a", "a", "a"]}))
        expected4 = pd.DataFrame({"key": ["key1"], "count": [4]})
        pdt.assert_frame_equal(ans4, expected4, check_like=True)

    def test_direct_access_to_value_state(self):
        processor = RunningCountStatefulProcessorFactory().row()
        tester = TwsTester(processor)
        tester.updateValueState("count", "foo", (5,))
        tester.test("foo", [Row(value="q")])
        self.assertEqual(tester.peekValueState("count", "foo"), (6,))

    def test_topk_processor(self):
        processor = TopKProcessorFactory(k=2).row()
        tester = TwsTester(processor)

        ans1 = tester.test("key1", [Row(score=2.0), Row(score=3.0), Row(score=1.0)])
        self.assertEqual(ans1, [Row(key="key1", score=3.0), Row(key="key1", score=2.0)])

        ans2 = tester.test(
            "key2", [Row(score=10.0), Row(score=20.0), Row(score=30.0), Row(score=40.0)]
        )
        self.assertEqual(ans2, [Row(key="key2", score=40.0), Row(key="key2", score=30.0)])

        ans3 = tester.test("key3", [Row(score=100.0)])
        self.assertEqual(ans3, [Row(key="key3", score=100.0)])

        self.assertEqual(tester.peekListState("topK", "key1"), [(3.0,), (2.0,)])
        self.assertEqual(tester.peekListState("topK", "key2"), [(40.0,), (30.0,)])
        self.assertEqual(tester.peekListState("topK", "key3"), [(100.0,)])
        self.assertEqual(tester.peekListState("topK", "key4"), [])

        ans4 = tester.test("key1", [Row(score=10.0)])
        self.assertEqual(ans4, [Row(key="key1", score=10.0), Row(key="key1", score=3.0)])
        self.assertEqual(tester.peekListState("topK", "key1"), [(10.0,), (3.0,)])

    def test_topk_processor_pandas(self):
        processor = TopKProcessorFactory(k=2).pandas()
        tester = TwsTester(processor)

        ans1 = tester.testInPandas("key1", pd.DataFrame({"score": [2.0, 3.0, 1.0]}))
        expected1 = pd.DataFrame({"key": ["key1", "key1"], "score": [3.0, 2.0]})
        pdt.assert_frame_equal(ans1, expected1, check_like=True)

        ans2 = tester.testInPandas("key2", pd.DataFrame({"score": [10.0, 20.0, 30.0, 40.0]}))
        expected2 = pd.DataFrame({"key": ["key2", "key2"], "score": [40.0, 30.0]})
        pdt.assert_frame_equal(ans2, expected2, check_like=True)

        ans3 = tester.testInPandas("key3", pd.DataFrame({"score": [100.0]}))
        expected3 = pd.DataFrame({"key": ["key3"], "score": [100.0]})
        pdt.assert_frame_equal(ans3, expected3, check_like=True)

        ans4 = tester.testInPandas("key1", pd.DataFrame({"score": [10.0]}))
        expected4 = pd.DataFrame({"key": ["key1", "key1"], "score": [10.0, 3.0]})
        pdt.assert_frame_equal(ans4, expected4, check_like=True)

    def test_direct_access_to_list_state(self):
        processor = TopKProcessorFactory(k=2).row()
        tester = TwsTester(processor)

        tester.updateListState("topK", "a", [(6.0,), (5.0,)])
        tester.updateListState("topK", "b", [(8.0,), (7.0,)])
        tester.test("a", [Row(score=10.0)])
        tester.test("b", [Row(score=7.5)])
        tester.test("c", [Row(score=1.0)])

        assert tester.peekListState("topK", "a") == [(10.0,), (6.0,)]
        assert tester.peekListState("topK", "b") == [(8.0,), (7.5,)]
        assert tester.peekListState("topK", "c") == [(1.0,)]
        assert tester.peekListState("topK", "d") == []

    def test_word_frequency_processor(self):
        processor = WordFrequencyProcessorFactory().row()
        tester = TwsTester(processor)

        ans1 = tester.test(
            "user1",
            [
                Row(word="hello"),
                Row(word="world"),
                Row(word="hello"),
                Row(word="world"),
            ],
        )
        self.assertEqual(
            ans1,
            [
                Row(key="user1", word="hello", count=1),
                Row(key="user1", word="world", count=1),
                Row(key="user1", word="hello", count=2),
                Row(key="user1", word="world", count=2),
            ],
        )

        ans2 = tester.test("user2", [Row(word="hello"), Row(word="spark")])
        self.assertEqual(
            ans2,
            [
                Row(key="user2", word="hello", count=1),
                Row(key="user2", word="spark", count=1),
            ],
        )

        # Check state using peekMapState.
        self.assertEqual(
            tester.peekMapState("frequencies", "user1"),
            {("hello",): (2,), ("world",): (2,)},
        )
        self.assertEqual(
            tester.peekMapState("frequencies", "user2"),
            {("hello",): (1,), ("spark",): (1,)},
        )
        self.assertEqual(tester.peekMapState("frequencies", "user3"), {})

        # Process more data for user1.
        ans3 = tester.test("user1", [Row(word="hello"), Row(word="test")])
        self.assertEqual(
            ans3,
            [
                Row(key="user1", word="hello", count=3),
                Row(key="user1", word="test", count=1),
            ],
        )
        self.assertEqual(
            tester.peekMapState("frequencies", "user1"),
            {("hello",): (3,), ("world",): (2,), ("test",): (1,)},
        )

    def test_word_frequency_processor_pandas(self):
        processor = WordFrequencyProcessorFactory().pandas()
        tester = TwsTester(processor)

        input_df1 = pd.DataFrame({"word": ["hello", "world", "hello", "world"]})
        ans1 = tester.testInPandas("user1", input_df1)
        expected1 = pd.DataFrame(
            {
                "key": ["user1", "user1", "user1", "user1"],
                "word": ["hello", "world", "hello", "world"],
                "count": [1, 1, 2, 2],
            }
        )
        pdt.assert_frame_equal(ans1, expected1, check_like=True)

        input_df2 = pd.DataFrame({"word": ["hello", "spark"]})
        ans2 = tester.testInPandas("user2", input_df2)
        expected2 = pd.DataFrame(
            {
                "key": ["user2", "user2"],
                "word": ["hello", "spark"],
                "count": [1, 1],
            }
        )
        pdt.assert_frame_equal(ans2, expected2, check_like=True)

        input_df3 = pd.DataFrame({"word": ["hello", "test"]})
        ans3 = tester.testInPandas("user1", input_df3)
        expected3 = pd.DataFrame(
            {
                "key": ["user1", "user1"],
                "word": ["hello", "test"],
                "count": [3, 1],
            }
        )
        pdt.assert_frame_equal(ans3, expected3, check_like=True)

    def test_direct_access_to_map_state(self):
        processor = WordFrequencyProcessorFactory().row()
        tester = TwsTester(processor)

        tester.updateMapState("frequencies", "user1", {("hello",): (5,), ("world",): (3,)})
        tester.updateMapState("frequencies", "user2", {("spark",): (10,)})

        tester.test("user1", [Row(word="hello"), Row(word="goodbye")])
        tester.test("user2", [Row(word="spark")])
        tester.test("user3", [Row(word="new")])

        self.assertEqual(
            tester.peekMapState("frequencies", "user1"),
            {("hello",): (6,), ("world",): (3,), ("goodbye",): (1,)},
        )
        self.assertEqual(tester.peekMapState("frequencies", "user2"), {("spark",): (11,)})
        self.assertEqual(tester.peekMapState("frequencies", "user3"), {("new",): (1,)})
        self.assertEqual(tester.peekMapState("frequencies", "user4"), {})

    # Example of how TwsTester can be used to test step function.
    def test_step_function(self):
        processor = RunningCountStatefulProcessorFactory().row()
        tester = TwsTester(processor)

        # Example of helper function using TwsTester to inspect how processing a single row changes
        # state.
        def step_function(key: str, input_row: str, state_in: int) -> int:
            tester.updateValueState("count", key, (state_in,))
            tester.test(key, [Row(value=input_row)])
            return tester.peekValueState("count", key)[0]

        self.assertEqual(step_function("key1", "a", 10), 11)

    # Example of how TwsTester can be used to simulate real-time mode (row by row processing).
    def test_row_by_row(self):
        processor = RunningCountStatefulProcessorFactory().row()
        tester = TwsTester(processor)

        # Example of helper function to test how TransformWithState processes rows one-by-one,
        # which can be used to simulate real-time mode.
        def test_row_by_row_helper(input_rows: list[tuple[str, Row]]) -> list[Row]:
            result: list[Row] = []
            for key, row in input_rows:
                result += tester.test(key, [row])
            return result

        ans = test_row_by_row_helper(
            [
                ("key1", Row(value="a")),
                ("key2", Row(value="b")),
                ("key1", Row(value="c")),
                ("key2", Row(value="b")),
                ("key1", Row(value="c")),
                ("key1", Row(value="c")),
                ("key3", Row(value="q")),
            ]
        )
        self.assertEqual(
            ans,
            [
                Row(key="key1", count=1),
                Row(key="key2", count=1),
                Row(key="key1", count=2),
                Row(key="key2", count=2),
                Row(key="key1", count=3),
                Row(key="key1", count=4),
                Row(key="key3", count=1),
            ],
        )

    # Tests that TwsTester calls handleInitialState.
    def test_initial_state_row(self):
        processor = RunningCountStatefulProcessorFactory().row()
        tester = TwsTester(
            processor,
            initialStateRow=[
                ("a", Row(initial_count=10)),
                ("b", Row(initial_count=20)),
            ],
        )
        self.assertEqual(tester.peekValueState("count", "a"), (10,))
        self.assertEqual(tester.peekValueState("count", "b"), (20,))

        ans1 = tester.test("a", [Row(value="a")])
        self.assertEqual(ans1, [Row(key="a", count=11)])

        ans2 = tester.test("c", [Row(value="c")])
        self.assertEqual(ans2, [Row(key="c", count=1)])

    def test_initial_state_pandas(self):
        processor = RunningCountStatefulProcessorFactory().pandas()
        tester = TwsTester(
            processor,
            initialStatePandas=[
                ("a", pd.DataFrame({"initial_count": [10]})),
                ("b", pd.DataFrame({"initial_count": [20]})),
            ],
        )
        self.assertEqual(tester.peekValueState("count", "a"), (10,))
        self.assertEqual(tester.peekValueState("count", "b"), (20,))

        ans1 = tester.testInPandas("a", pd.DataFrame({"value": ["a"]}))
        expected1 = pd.DataFrame({"key": ["a"], "count": [11]})
        pdt.assert_frame_equal(ans1, expected1, check_like=True)

        ans2 = tester.testInPandas("c", pd.DataFrame({"value": ["c"]}))
        expected2 = pd.DataFrame({"key": ["c"], "count": [1]})
        pdt.assert_frame_equal(ans2, expected2, check_like=True)

    def test_all_methods_processor(self):
        """Test that TwsTester exercises all state methods."""
        processor = AllMethodsTestProcessorFactory().row()
        tester = TwsTester(processor)

        results = tester.test(
            "k",
            [
                Row(cmd="value-exists"),  # false
                Row(cmd="value-set"),  # set to 42
                Row(cmd="value-exists"),  # true
                Row(cmd="value-clear"),  # clear
                Row(cmd="value-exists"),  # false again
                Row(cmd="list-exists"),  # false
                Row(cmd="list-append"),  # append a, b
                Row(cmd="list-exists"),  # true
                Row(cmd="list-append-array"),  # append c, d
                Row(cmd="list-get"),  # a,b,c,d
                Row(cmd="map-exists"),  # false
                Row(cmd="map-add"),  # add x=1, y=2, z=3
                Row(cmd="map-exists"),  # true
                Row(cmd="map-keys"),  # x,y,z
                Row(cmd="map-values"),  # 1,2,3
                Row(cmd="map-iterator"),  # x=1,y=2,z=3
                Row(cmd="map-remove"),  # remove y
                Row(cmd="map-keys"),  # x,z
                Row(cmd="map-clear"),  # clear map
                Row(cmd="map-exists"),  # false
            ],
        )

        self.assertEqual(
            results,
            [
                Row(key="k", result="value-exists:False"),
                Row(key="k", result="value-set:done"),
                Row(key="k", result="value-exists:True"),
                Row(key="k", result="value-clear:done"),
                Row(key="k", result="value-exists:False"),
                Row(key="k", result="list-exists:False"),
                Row(key="k", result="list-append:done"),
                Row(key="k", result="list-exists:True"),
                Row(key="k", result="list-append-array:done"),
                Row(key="k", result="list-get:a,b,c,d"),
                Row(key="k", result="map-exists:False"),
                Row(key="k", result="map-add:done"),
                Row(key="k", result="map-exists:True"),
                Row(key="k", result="map-keys:x,y,z"),
                Row(key="k", result="map-values:1,2,3"),
                Row(key="k", result="map-iterator:x=1,y=2,z=3"),
                Row(key="k", result="map-remove:done"),
                Row(key="k", result="map-keys:x,z"),
                Row(key="k", result="map-clear:done"),
                Row(key="k", result="map-exists:False"),
            ],
        )

    def test_both_initial_states_specified(self):
        processor = RunningCountStatefulProcessorFactory().row()
        with self.assertRaises(AssertionError):
            TwsTester(
                processor,
                initialStateRow=[("a", Row(initial_count=10))],
                initialStatePandas=[("a", pd.DataFrame({"initial_count": [10]}))],
            )

    def test_timer_registration_raises_error_in_none_mode(self):
        processor = RunningCountStatefulProcessorFactory().row()
        tester = TwsTester(processor, timeMode="None")
        with self.assertRaisesRegex(PySparkValueError, "UNSUPPORTED_OPERATION"):
            tester.handle.registerTimer(12345)

    def test_delete_timer_raises_error_in_none_mode(self):
        processor = RunningCountStatefulProcessorFactory().row()
        tester = TwsTester(processor, timeMode="None")
        with self.assertRaisesRegex(PySparkValueError, "UNSUPPORTED_OPERATION"):
            tester.handle.deleteTimer(12345)

    def test_list_timers_raises_error_in_none_mode(self):
        processor = RunningCountStatefulProcessorFactory().row()
        tester = TwsTester(processor, timeMode="None")
        with self.assertRaisesRegex(PySparkValueError, "UNSUPPORTED_OPERATION"):
            list(tester.handle.listTimers())

    def test_empty_input_row(self):
        processor = RunningCountStatefulProcessorFactory().row()
        tester = TwsTester(processor)
        result = tester.test("key1", [])
        self.assertEqual(result, [Row(key="key1", count=0)])

    def test_empty_input_pandas(self):
        processor = RunningCountStatefulProcessorFactory().pandas()
        tester = TwsTester(processor)
        input_df = pd.DataFrame({"value": []})
        result = tester.testInPandas("key1", input_df)
        expected = pd.DataFrame({"key": ["key1"], "count": [0]})
        pdt.assert_frame_equal(result, expected, check_like=True)

    def test_empty_initial_state_row(self):
        processor = RunningCountStatefulProcessorFactory().row()
        tester = TwsTester(processor, initialStateRow=[])
        result = tester.test("key1", [Row(value="a")])
        self.assertEqual(result, [Row(key="key1", count=1)])

    def test_empty_initial_state_pandas(self):
        processor = RunningCountStatefulProcessorFactory().pandas()
        tester = TwsTester(processor, initialStatePandas=[])
        result = tester.testInPandas("key1", pd.DataFrame({"value": ["a"]}))
        expected = pd.DataFrame({"key": ["key1"], "count": [1]})
        pdt.assert_frame_equal(result, expected, check_like=True)

    def test_none_values_in_input(self):
        processor = RunningCountStatefulProcessorFactory().row()
        tester = TwsTester(processor)
        result = tester.test("key1", [Row(value=None), Row(value="a")])
        self.assertEqual(result, [Row(key="key1", count=2)])

    def test_single_row_input(self):
        processor = RunningCountStatefulProcessorFactory().row()
        tester = TwsTester(processor)
        result = tester.test("key1", [Row(value="a")])
        self.assertEqual(result, [Row(key="key1", count=1)])

    def test_peek_nonexistent_state(self):
        processor = RunningCountStatefulProcessorFactory().row()
        tester = TwsTester(processor)
        result = tester.peekValueState("count", "nonexistent_key")
        self.assertIsNone(result)

    def test_update_nonexistent_state(self):
        processor = RunningCountStatefulProcessorFactory().row()
        tester = TwsTester(processor)
        with self.assertRaises(AssertionError):
            tester.updateValueState("nonexistent_state", "key1", (5,))

    def test_peek_state_before_initialization(self):
        processor = RunningCountStatefulProcessorFactory().row()
        tester = TwsTester(processor)
        with self.assertRaises(AssertionError):
            tester.peekValueState("nonexistent_state", "key1")

    def test_clear_nonexistent_value_state(self):
        processor = RunningCountStatefulProcessorFactory().row()
        tester = TwsTester(processor)
        tester.handle.setGroupingKey("key1")
        value_state = tester.handle.getValueState("count", "int")
        value_state.clear()

    def test_clear_nonexistent_list_state(self):
        processor = TopKProcessorFactory(k=2).row()
        tester = TwsTester(processor)
        tester.handle.setGroupingKey("key1")
        list_state = tester.handle.getListState("topK", "double")
        list_state.clear()

    def test_clear_nonexistent_map_state(self):
        processor = WordFrequencyProcessorFactory().row()
        tester = TwsTester(processor)
        tester.handle.setGroupingKey("key1")
        map_state = tester.handle.getMapState("frequencies", "string", "int")
        map_state.clear()

    def test_delete_if_exists_nonexistent(self):
        processor = RunningCountStatefulProcessorFactory().row()
        tester = TwsTester(processor)
        tester.handle.deleteIfExists("nonexistent_state")

    def test_numeric_key(self):
        processor = RunningCountStatefulProcessorFactory().row()
        tester = TwsTester(processor)
        self.assertEqual(tester.test(1, [Row(value="a"), Row(value="c")]), [Row(key=1, count=2)])
        self.assertEqual(tester.test(2, [Row(value="b")]), [Row(key=2, count=1)])

    def test_tuple_key(self):
        processor = RunningCountStatefulProcessorFactory().row()
        tester = TwsTester(processor)
        result = tester.test((1, 2), [Row(value="a"), Row(value="b")])
        self.assertEqual(result, [Row(key=(1, 2), count=2)])

    def test_special_characters_in_key(self):
        processor = RunningCountStatefulProcessorFactory().row()
        tester = TwsTester(processor)
        result = tester.test("key@#$%", [Row(value="a"), Row(value="b")])
        self.assertEqual(result, [Row(key="key@#$%", count=2)])

    def test_null_key_value_row(self):
        processor = RunningCountStatefulProcessorFactory().row()
        tester = TwsTester(processor)
        result = tester.test(None, [Row(value="a"), Row(value="b")])
        self.assertEqual(result, [Row(key=None, count=2)])

    def test_null_key_value_pandas(self):
        processor = RunningCountStatefulProcessorFactory().pandas()
        tester = TwsTester(processor)
        input_df = pd.DataFrame({"value": ["a", "b"]})
        result = tester.testInPandas(None, input_df)
        expected = pd.DataFrame({"key": [None], "count": [2]})
        pdt.assert_frame_equal(result, expected, check_like=True)

    def test_pandas_with_multiple_rows(self):
        processor = RunningCountStatefulProcessorFactory().pandas()
        tester = TwsTester(processor)
        input_df = pd.DataFrame({"value": ["a", "b", "c"]})
        result = tester.testInPandas("key1", input_df)
        expected = pd.DataFrame({"key": ["key1"], "count": [3]})
        pdt.assert_frame_equal(result, expected, check_like=True)

    def test_pandas_dtype_preservation(self):
        processor = RunningCountStatefulProcessorFactory().pandas()
        tester = TwsTester(processor)
        input_df = pd.DataFrame({"value": ["a"]})
        result = tester.testInPandas("key1", input_df)
        self.assertEqual(result["count"].dtype, "int64")

    def test_processor_init_called_once(self):
        from pyspark.sql.streaming.stateful_processor import (
            StatefulProcessor,
            StatefulProcessorHandle,
        )
        from typing import Iterator

        init_call_count = [0]

        class InitCountingProcessor(StatefulProcessor):
            def init(self, handle: StatefulProcessorHandle) -> None:
                init_call_count[0] += 1
                self.handle = handle
                self.state = handle.getValueState("count", "int")

            def handleInputRows(self, key, rows, timerValues) -> Iterator:
                val = self.state.get() if self.state.exists() else None
                count = val[0] if val else 0
                for row in rows:
                    count += 1
                self.state.update((count,))
                yield Row(key=key[0], count=count)

        processor = InitCountingProcessor()
        tester = TwsTester(processor)
        tester.test("key1", [Row(value="a")])
        tester.test("key1", [Row(value="b")])
        self.assertEqual(init_call_count[0], 1)

    def test_delete_value_state(self):
        """Test that deleteState correctly deletes value state for a given key."""
        processor = RunningCountStatefulProcessorFactory().row()
        tester = TwsTester(processor)

        tester.updateValueState("count", "key1", (10,))
        tester.updateValueState("count", "key2", (20,))
        self.assertEqual(tester.peekValueState("count", "key1"), (10,))

        tester.deleteState("count", "key1")
        self.assertIsNone(tester.peekValueState("count", "key1"))
        self.assertEqual(tester.peekValueState("count", "key2"), (20,))

    def test_delete_list_state(self):
        """Test that deleteState correctly deletes list state for a given key."""
        processor = TopKProcessorFactory(k=3).row()
        tester = TwsTester(processor)

        tester.updateListState("topK", "key1", [(1.0,), (2.0,), (3.0,)])
        tester.updateListState("topK", "key2", [(4.0,), (5.0,)])
        self.assertEqual(tester.peekListState("topK", "key1"), [(1.0,), (2.0,), (3.0,)])

        tester.deleteState("topK", "key1")
        self.assertEqual(tester.peekListState("topK", "key1"), [])
        self.assertEqual(tester.peekListState("topK", "key2"), [(4.0,), (5.0,)])

    def test_delete_map_state(self):
        """Test that deleteState correctly deletes map state for a given key."""
        processor = WordFrequencyProcessorFactory().row()
        tester = TwsTester(processor)

        tester.updateMapState("frequencies", "user1", {("hello",): (5,), ("world",): (3,)})
        tester.updateMapState("frequencies", "user2", {("spark",): (10,)})
        self.assertEqual(
            tester.peekMapState("frequencies", "user1"),
            {("hello",): (5,), ("world",): (3,)},
        )

        tester.deleteState("frequencies", "user1")
        self.assertEqual(tester.peekMapState("frequencies", "user1"), {})
        self.assertEqual(tester.peekMapState("frequencies", "user2"), {("spark",): (10,)})

    def test_delete_nonexistent_state_raises_error(self):
        """Test that deleteState raises an error for non-existent state."""
        processor = RunningCountStatefulProcessorFactory().row()
        tester = TwsTester(processor)

        with self.assertRaisesRegex(PySparkAssertionError, "STATE_NOT_EXISTS"):
            tester.deleteState("nonexistent_state", "key1")

    # Timer tests.

    def test_processing_time_timers(self):
        """Test that TwsTester supports ProcessingTime timers."""
        processor = SessionTimeoutProcessorFactory().row()
        tester = TwsTester(processor, timeMode="ProcessingTime")

        # Process input for key1 - should register a timer at t=10000.
        result1 = tester.test("key1", [Row(value="hello")])
        self.assertEqual(result1, [Row(key="key1", result="received:hello")])

        # Set processing time to 5000 - timer should NOT fire yet.
        expired1 = tester.setProcessingTime(5000)
        self.assertEqual(expired1, [])

        # Process input for key2 at t=5000 - should register timer at t=15000.
        result2 = tester.test("key2", [Row(value="world")])
        self.assertEqual(result2, [Row(key="key2", result="received:world")])

        # Set processing time to 11000 - key1's timer should fire.
        expired2 = tester.setProcessingTime(11000)
        self.assertEqual(expired2, [Row(key="key1", result="session-expired")])

        # Set processing time to 16000 - key2's timer should fire.
        expired3 = tester.setProcessingTime(16000)
        self.assertEqual(expired3, [Row(key="key2", result="session-expired")])

        # Verify state is cleared after session expiry.
        self.assertIsNone(tester.peekValueState("lastSeen", "key1"))
        self.assertIsNone(tester.peekValueState("lastSeen", "key2"))

    def test_processing_time_timers_pandas(self):
        """Test that TwsTester supports ProcessingTime timers in Pandas mode."""
        processor = SessionTimeoutProcessorFactory().pandas()
        tester = TwsTester(processor, timeMode="ProcessingTime")

        # Process input for key1 - should register a timer at t=10000.
        result1 = tester.testInPandas("key1", pd.DataFrame({"value": ["hello"]}))
        expected1 = pd.DataFrame({"key": ["key1"], "result": ["received:hello"]})
        pdt.assert_frame_equal(result1, expected1, check_like=True)

        # Set processing time to 5000 - timer should NOT fire yet.
        expired1 = tester.setProcessingTime(5000)
        self.assertEqual(len(expired1), 0)

        # Process input for key2 at t=5000 - should register timer at t=15000.
        result2 = tester.testInPandas("key2", pd.DataFrame({"value": ["world"]}))
        expected2 = pd.DataFrame({"key": ["key2"], "result": ["received:world"]})
        pdt.assert_frame_equal(result2, expected2, check_like=True)

        # Set processing time to 11000 - key1's timer should fire.
        expired2 = tester.setProcessingTime(11000)
        expected_expired2 = pd.DataFrame({"key": ["key1"], "result": ["session-expired"]})
        pdt.assert_frame_equal(expired2, expected_expired2, check_like=True)

        # Set processing time to 16000 - key2's timer should fire.
        expired3 = tester.setProcessingTime(16000)
        expected_expired3 = pd.DataFrame({"key": ["key2"], "result": ["session-expired"]})
        pdt.assert_frame_equal(expired3, expected_expired3, check_like=True)

        # Verify state is cleared after session expiry.
        self.assertIsNone(tester.peekValueState("lastSeen", "key1"))
        self.assertIsNone(tester.peekValueState("lastSeen", "key2"))

    def test_event_time_timers_manual_watermark(self):
        """Test that TwsTester supports EventTime timers fired by manual watermark advance."""
        processor = EventTimeSessionProcessorFactory().row()

        def event_time_extractor(row: Row) -> int:
            return row.event_time_ms

        tester = TwsTester(
            processor,
            timeMode="EventTime",
            eventTimeExtractor=event_time_extractor,
        )

        # Process event at t=10000 for key1 - registers timer at t=15000.
        result1 = tester.test("key1", [Row(event_time_ms=10000, value="hello")])
        self.assertEqual(result1, [Row(key="key1", result="received:hello@10000")])

        # Process event at t=11000 for key2 - registers timer at t=16000.
        result2 = tester.test("key2", [Row(event_time_ms=11000, value="world")])
        self.assertEqual(result2, [Row(key="key2", result="received:world@11000")])

        # Set watermark to 15000 - key1's timer should fire.
        expired1 = tester.setWatermark(15000)
        self.assertEqual(len(expired1), 1)
        self.assertEqual(expired1[0], Row(key="key1", result="session-expired@watermark=15000"))

        # Set watermark to 16000 - key2's timer should fire.
        expired2 = tester.setWatermark(16000)
        self.assertEqual(len(expired2), 1)
        self.assertEqual(expired2[0], Row(key="key2", result="session-expired@watermark=16000"))

        # Verify state is cleared.
        self.assertIsNone(tester.peekValueState("lastEventTime", "key1"))
        self.assertIsNone(tester.peekValueState("lastEventTime", "key2"))

    def test_late_event_filtering(self):
        """Test that TwsTester filters late events in EventTime mode."""
        processor = EventTimeCountProcessorFactory().row()

        def event_time_extractor(row: Row) -> int:
            return row.event_time_ms

        tester = TwsTester(
            processor,
            timeMode="EventTime",
            eventTimeExtractor=event_time_extractor,
        )

        # Initially watermark is 0, so all events should be processed.
        result1 = tester.test(
            "key1",
            [
                Row(event_time_ms=5000, value="a"),
                Row(event_time_ms=6000, value="b"),
                Row(event_time_ms=7000, value="c"),
            ],
        )
        self.assertEqual(result1, [Row(key="key1", count=3)])
        self.assertEqual(tester.peekValueState("count", "key1"), (3,))

        # Set watermark to 6000.
        tester.setWatermark(6000)

        # Now watermark is 6000. Send events with mixed event times:
        # - (4000, "late1") -> event time 4000 <= watermark 6000, should be FILTERED.
        # - (5000, "late2") -> event time 5000 <= watermark 6000, should be FILTERED.
        # - (6000, "late3") -> event time 6000 <= watermark 6000, should be FILTERED.
        # - (8000, "ontime") -> event time 8000 > watermark 6000, should be PROCESSED.
        result2 = tester.test(
            "key1",
            [
                Row(event_time_ms=4000, value="late1"),
                Row(event_time_ms=5000, value="late2"),
                Row(event_time_ms=6000, value="late3"),
                Row(event_time_ms=8000, value="ontime"),
            ],
        )
        # Only 1 event should be processed (the on-time one)
        self.assertEqual(result2, [Row(key="key1", count=4)])  # 3 + 1 = 4
        self.assertEqual(tester.peekValueState("count", "key1"), (4,))

    def test_registers_on_time_events(self):
        "TwsTester with eventTimeExtractor should allow timer registration from on-time events"
        # With eventTimeExtractor, late events are filtered BEFORE reaching the processor,
        # so timers can only be registered from on-time events (no exception).
        tester = TwsTester(
            EventTimeSessionProcessorFactory().row(),
            timeMode="EventTime",
            eventTimeExtractor=lambda row: row.event_time_ms,
        )

        # Set watermark to 20 seconds.
        tester.setWatermark(20000)

        # Process a mix of late and on-time events.
        # Late event (10000) is filtered out, on-time event (25000) is processed.
        # Timer registered at 25000 + 5000 = 30000, which is > watermark, so no exception.
        result = tester.test(
            "key1",
            [Row(event_time_ms=10000, value="late"), Row(event_time_ms=25000, value="ontime")],
        )
        # Only the on-time event produces output.
        self.assertEqual(result, [Row(key="key1", result="received:ontime@25000")])

    def test_timer_registration_in_none_mode_raises_error(self):
        """Test that timer operations raise error in TimeMode.None."""
        processor = RunningCountStatefulProcessorFactory().row()
        tester = TwsTester(processor, timeMode="None")

        with self.assertRaisesRegex(PySparkValueError, "UNSUPPORTED_OPERATION"):
            tester.handle.registerTimer(12345)

    def test_set_processing_time_in_wrong_mode_raises_error(self):
        """Test that setProcessingTime raises error in non-ProcessingTime mode."""
        processor = RunningCountStatefulProcessorFactory().row()

        # Test in None mode.
        tester_none = TwsTester(processor, timeMode="None")
        with self.assertRaisesRegex(PySparkValueError, "UNSUPPORTED_OPERATION"):
            tester_none.setProcessingTime(1000)

    def test_set_watermark_in_wrong_mode_raises_error(self):
        """Test that setWatermark raises error in non-EventTime mode."""
        processor = RunningCountStatefulProcessorFactory().row()

        # Test in None mode.
        tester_none = TwsTester(processor, timeMode="None")
        with self.assertRaisesRegex(PySparkValueError, "UNSUPPORTED_OPERATION"):
            tester_none.setWatermark(1000)

        # Test in ProcessingTime mode.
        tester_pt = TwsTester(processor, timeMode="ProcessingTime")
        with self.assertRaisesRegex(PySparkValueError, "UNSUPPORTED_OPERATION"):
            tester_pt.setWatermark(1000)

    def test_eventtime_mode_requires_extractor(self):
        """Test that EventTime mode requires eventTimeExtractor."""
        processor = RunningCountStatefulProcessorFactory().row()

        with self.assertRaisesRegex(PySparkValueError, "ARGUMENT_REQUIRED"):
            TwsTester(processor, timeMode="EventTime")

    def test_processing_time_must_move_forward(self):
        """Test that setProcessingTime requires time to move forward."""
        processor = RunningCountStatefulProcessorFactory().row()
        tester = TwsTester(processor, timeMode="ProcessingTime")

        tester.setProcessingTime(1000)

        # Same time should raise error.
        with self.assertRaises(IllegalArgumentException):
            tester.setProcessingTime(1000)

        # Earlier time should raise error.
        with self.assertRaises(IllegalArgumentException):
            tester.setProcessingTime(500)

    def test_watermark_must_move_forward(self):
        """Test that setWatermark requires watermark to move forward."""
        processor = RunningCountStatefulProcessorFactory().row()

        def event_time_extractor(row: Row) -> int:
            return row.event_time_ms

        tester = TwsTester(
            processor,
            timeMode="EventTime",
            eventTimeExtractor=event_time_extractor,
        )

        tester.setWatermark(1000)

        # Same watermark should raise error.
        with self.assertRaises(IllegalArgumentException):
            tester.setWatermark(1000)

        # Earlier watermark should raise error.
        with self.assertRaises(IllegalArgumentException):
            tester.setWatermark(500)

    def test_timer_cannot_be_registered_in_past(self):
        """Test that timer cannot be registered at or before current watermark in EventTime mode."""
        processor = EventTimeSessionProcessorFactory().row()

        def event_time_extractor(row: Row) -> int:
            return row.event_time_ms

        tester = TwsTester(
            processor,
            timeMode="EventTime",
            eventTimeExtractor=event_time_extractor,
        )

        # Set watermark to 10000.
        tester.setWatermark(10000)

        # Try to register a timer at or before watermark - should raise error.
        tester.handle.setGroupingKey("key1")
        with self.assertRaises(IllegalArgumentException):
            tester.handle.registerTimer(10000)  # exactly at watermark

        with self.assertRaises(IllegalArgumentException):
            tester.handle.registerTimer(5000)  # before watermark


@unittest.skipIf(
    not have_pandas or not have_pyarrow or os.environ.get("PYTHON_GIL", "?") == "0",
    pandas_requirement_message or pyarrow_requirement_message or "Not supported in no-GIL mode",
)
class TwsTesterFuzzTests(ReusedSQLTestCase):
    @classmethod
    def conf(cls):
        cfg = SparkConf()
        cfg.set("spark.sql.shuffle.partitions", "1")
        cfg.set(
            "spark.sql.streaming.stateStore.providerClass",
            "org.apache.spark.sql.execution.streaming.state.RocksDBStateStoreProvider",
        )
        return cfg

    # Helper that runs real streaming query with TWS and compares results with TwsTester output.
    # Supports both row mode (use_pandas=False) and Pandas mode (use_pandas=True).
    # Supports multiple batches.
    # Assumes that the first column is the key column.
    # Output rows are checked ignoring order.
    # Correct only if results do not depend on reordering rows within batch.
    def _check_tws_tester(
        self,
        processor: StatefulProcessor,
        batches: list[list[tuple]],
        input_columns: list[str],
        input_types: list[str],
        output_schema: StructType,
        use_pandas: bool = False,
    ):
        num_batches = len(batches)
        key_column: str = input_columns[0]  # Assume first column is the key column.
        value_columns: list[str] = input_columns[1:]  # All columns except the key column.

        num_input_columns: int = len(input_columns)
        assert len(input_types) == num_input_columns

        # Use TwsTester to compute expected results.
        tester = TwsTester(processor)

        def _run_tester(batch: list[tuple]) -> list[Row]:
            # Group batch by key.
            grouped = {}
            for row in batch:
                key = row[0]
                if key not in grouped:
                    grouped[key] = []
                grouped[key].append(row[1:])

            results: list[Row] = []
            for key, values in grouped.items():
                if use_pandas:
                    input_to_tester = pd.DataFrame(values, columns=value_columns)
                    result_df = tester.testInPandas(key, input_to_tester)
                    results.extend(self.spark.createDataFrame(result_df).collect())
                else:
                    input_to_tester: list[Row] = [
                        Row(**dict(zip(value_columns, v))) for v in values
                    ]
                    results.extend(tester.test(key, input_to_tester))
            return results

        expected_results: list[list[Row]] = [_run_tester(batch) for batch in batches]

        # Create streaming DataFrame, it will read input from files.
        input_path = tempfile.mkdtemp()
        df = self.spark.readStream.format("text").load(input_path)
        df_split = df.withColumn("split_values", split(df["value"], ","))
        df_final = df_split.select(
            *[
                df_split.split_values.getItem(i).alias(col_name).cast(input_types[i])
                for i, col_name in enumerate(input_columns)
            ]
        )
        self.assertTrue(df_final.isStreaming)

        # Apply TransformWithState.
        df_grouped = df_final.groupBy(key_column)
        if use_pandas:
            tws_df = df_grouped.transformWithStateInPandas(
                statefulProcessor=processor,
                outputStructType=output_schema,
                outputMode="Update",
                timeMode="None",
            )
        else:
            tws_df = df_grouped.transformWithState(
                statefulProcessor=processor,
                outputStructType=output_schema,
                outputMode="Update",
                timeMode="None",
            )

        # Start streaming query collecting TWS results.
        actual_results: list[list[Row]] = []

        def _collect_results(batch_df: DataFrame, batch_id):
            rows: list[Row] = batch_df.collect()
            actual_results.append(rows)

        query: StreamingQuery = (
            tws_df.writeStream.queryName("test_tmp_query")
            .foreachBatch(_collect_results)
            .outputMode("update")
            .start()
        )

        # Process each input batch.
        def _write_batch(batch_id: int, batch: list[tuple]):
            with open(input_path + f"/test{batch_id}.txt", "w") as f:
                for row in batch:
                    assert len(row) == num_input_columns
                    f.write(",".join(str(value) for value in row) + "\n")

        try:
            for batch_id, batch in enumerate(batches):
                _write_batch(batch_id, batch)
                query.processAllAvailable()
            query.awaitTermination(1)
            self.assertTrue(query.exception() is None)
        finally:
            query.stop()

        # Assert actual results are equal to expected results.
        self.assertEqual(len(actual_results), num_batches)
        self.assertEqual(len(expected_results), num_batches)
        for batch_id in range(num_batches):
            # This checks that collections are the same up to order, not just compares lengths.
            self.assertCountEqual(actual_results[batch_id], expected_results[batch_id])

    def test_fuzz_running_count(self):
        output_schema = StructType(
            [
                StructField("key", StringType(), True),
                StructField("count", IntegerType(), True),
            ]
        )
        batches: list[list[tuple[str, str]]] = []
        for batch_id in range(5):
            batch: list[tuple[str, str]] = []
            for _ in range(200):
                key: str = f"key{random.randint(0, 9)}"
                value: str = "".join(random.choices("abcdefghijklmnopqrstuvwxyz", k=5))
                batch.append((key, value))
            batches.append(batch)
        input_columns: list[str] = ["key", "value"]
        input_types: list[str] = ["string", "string"]

        for use_pandas in [False, True]:
            self._check_tws_tester(
                RunningCountStatefulProcessorFactory().get(use_pandas=use_pandas),
                batches,
                input_columns,
                input_types,
                output_schema,
                use_pandas=use_pandas,
            )

    def test_fuzz_topk(self):
        output_schema = StructType(
            [
                StructField("key", StringType(), True),
                StructField("score", DoubleType(), True),
            ]
        )
        batches: list[list[tuple[str, float]]] = []
        for batch_id in range(5):
            batch: list[tuple[str, float]] = []
            for _ in range(200):
                key: str = f"key{random.randint(0, 9)}"
                score: float = random.randint(0, 1000000) / 1000
                batch.append((key, score))
            batches.append(batch)
        input_columns: list[str] = ["key", "score"]
        input_types: list[str] = ["string", "double"]

        for use_pandas in [False, True]:
            self._check_tws_tester(
                TopKProcessorFactory(k=10).get(use_pandas=use_pandas),
                batches,
                input_columns,
                input_types,
                output_schema,
                use_pandas=use_pandas,
            )

    def test_fuzz_word_frequency(self):
        output_schema = StructType(
            [
                StructField("key", StringType(), True),
                StructField("word", StringType(), True),
                StructField("count", IntegerType(), True),
            ]
        )
        batches: list[list[tuple[str, str]]] = []
        for batch_id in range(5):
            batch: list[tuple[str, str]] = []
            for _ in range(200):
                key = f"key{random.randint(0, 9)}"
                word = "".join(random.choices("abcdefghijklmnopqrstuvwxyz", k=1))
                batch.append((key, word))
            batches.append(batch)
        input_columns: list[str] = ["key", "word"]
        input_types: list[str] = ["string", "string"]

        for use_pandas in [False, True]:
            self._check_tws_tester(
                WordFrequencyProcessorFactory().get(use_pandas=use_pandas),
                batches,
                input_columns,
                input_types,
                output_schema,
                use_pandas=use_pandas,
            )


if __name__ == "__main__":
    from pyspark.testing import main

    main()



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from pyspark.sql import Row
from pyspark.testing.sqlutils import ReusedSQLTestCase


class SQLTestsMixin:
    def test_simple(self):
        res = self.spark.sql("SELECT 1 + 1").collect()
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0][0], 2)

    def test_args_dict(self):
        with self.temp_view("test"):
            self.spark.range(10).createOrReplaceTempView("test")
            df = self.spark.sql(
                "SELECT * FROM IDENTIFIER(:table_name)",
                args={"table_name": "test"},
            )

            self.assertEqual(df.count(), 10)
            self.assertEqual(df.limit(5).count(), 5)
            self.assertEqual(df.offset(5).count(), 5)

            self.assertEqual(df.take(1), [Row(id=0)])
            self.assertEqual(df.tail(1), [Row(id=9)])

    def test_args_list(self):
        with self.temp_view("test"):
            self.spark.range(10).createOrReplaceTempView("test")
            df = self.spark.sql(
                "SELECT * FROM test WHERE ? < id AND id < ?",
                args=[1, 6],
            )

            self.assertEqual(df.count(), 4)
            self.assertEqual(df.limit(3).count(), 3)
            self.assertEqual(df.offset(3).count(), 1)

            self.assertEqual(df.take(1), [Row(id=2)])
            self.assertEqual(df.tail(1), [Row(id=5)])

    def test_kwargs_literal(self):
        with self.temp_view("test"):
            self.spark.range(10).createOrReplaceTempView("test")

            df = self.spark.sql(
                "SELECT * FROM IDENTIFIER(:table_name) WHERE {m1} < id AND id < {m2} OR id = {m3}",
                args={"table_name": "test"},
                m1=3,
                m2=7,
                m3=9,
            )

            self.assertEqual(df.count(), 4)
            self.assertEqual(df.collect(), [Row(id=4), Row(id=5), Row(id=6), Row(id=9)])
            self.assertEqual(df.take(1), [Row(id=4)])
            self.assertEqual(df.tail(1), [Row(id=9)])

    def test_kwargs_literal_multiple_ref(self):
        with self.temp_view("test"):
            self.spark.range(10).createOrReplaceTempView("test")

            df = self.spark.sql(
                "SELECT * FROM IDENTIFIER(:table_name) WHERE {m} = id OR id > {m} OR {m} < 0",
                args={"table_name": "test"},
                m=6,
            )

            self.assertEqual(df.count(), 4)
            self.assertEqual(df.collect(), [Row(id=6), Row(id=7), Row(id=8), Row(id=9)])
            self.assertEqual(df.take(1), [Row(id=6)])
            self.assertEqual(df.tail(1), [Row(id=9)])

    def test_kwargs_dataframe(self):
        df0 = self.spark.range(10)
        df1 = self.spark.sql(
            "SELECT * FROM {df} WHERE id > 4",
            df=df0,
        )

        self.assertEqual(df0.schema, df1.schema)
        self.assertEqual(df1.count(), 5)
        self.assertEqual(df1.take(1), [Row(id=5)])
        self.assertEqual(df1.tail(1), [Row(id=9)])

    def test_kwargs_dataframe_with_column(self):
        df0 = self.spark.range(10)
        df1 = self.spark.sql(
            "SELECT * FROM {df} WHERE {df.id} > :m1 AND {df[id]} < :m2",
            {"m1": 4, "m2": 9},
            df=df0,
        )

        self.assertEqual(df0.schema, df1.schema)
        self.assertEqual(df1.count(), 4)
        self.assertEqual(df1.take(1), [Row(id=5)])
        self.assertEqual(df1.tail(1), [Row(id=8)])

    def test_nested_view(self):
        with self.temp_view("v1", "v2", "v3", "v4"):
            self.spark.range(10).createOrReplaceTempView("v1")
            self.spark.sql(
                "SELECT * FROM IDENTIFIER(:view) WHERE id > :m",
                args={"view": "v1", "m": 1},
            ).createOrReplaceTempView("v2")
            self.spark.sql(
                "SELECT * FROM IDENTIFIER(:view) WHERE id > :m",
                args={"view": "v2", "m": 2},
            ).createOrReplaceTempView("v3")
            self.spark.sql(
                "SELECT * FROM IDENTIFIER(:view) WHERE id > :m",
                args={"view": "v3", "m": 3},
            ).createOrReplaceTempView("v4")

            df = self.spark.sql("select * from v4")
            self.assertEqual(df.count(), 6)
            self.assertEqual(df.take(1), [Row(id=4)])
            self.assertEqual(df.tail(1), [Row(id=9)])

    def test_nested_dataframe(self):
        df0 = self.spark.range(10)
        df1 = self.spark.sql(
            "SELECT * FROM {df} WHERE id > ?",
            args=[1],
            df=df0,
        )
        df2 = self.spark.sql(
            "SELECT * FROM {df} WHERE id > ?",
            args=[2],
            df=df1,
        )
        df3 = self.spark.sql(
            "SELECT * FROM {df} WHERE id > ?",
            args=[3],
            df=df2,
        )

        self.assertEqual(df0.schema, df1.schema)
        self.assertEqual(df1.count(), 8)
        self.assertEqual(df1.take(1), [Row(id=2)])
        self.assertEqual(df1.tail(1), [Row(id=9)])

        self.assertEqual(df0.schema, df2.schema)
        self.assertEqual(df2.count(), 7)
        self.assertEqual(df2.take(1), [Row(id=3)])
        self.assertEqual(df2.tail(1), [Row(id=9)])

        self.assertEqual(df0.schema, df3.schema)
        self.assertEqual(df3.count(), 6)
        self.assertEqual(df3.take(1), [Row(id=4)])
        self.assertEqual(df3.tail(1), [Row(id=9)])

    def test_lit_time(self):
        import datetime

        actual = self.spark.sql("select TIME '12:34:56'").first()[0]
        self.assertEqual(actual, datetime.time(12, 34, 56))


class SQLTests(SQLTestsMixin, ReusedSQLTestCase):
    pass


if __name__ == "__main__":
    from pyspark.testing import main

    main()



# -*- encoding: utf-8 -*-
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
import json
import tempfile
from io import StringIO
from pyspark.errors import ArithmeticException
from pyspark.logger.logger import PySparkLogger, SPARK_LOG_SCHEMA
from pyspark.sql import Row, functions as sf
from pyspark.testing import assertDataFrameEqual
from pyspark.testing.sqlutils import ReusedSQLTestCase


class LoggerTestsMixin:
    def setUp(self):
        self.handler = logging.StreamHandler(StringIO())

        self.logger = PySparkLogger.getLogger("TestLogger")
        self.logger.setLevel(logging.INFO)
        self.logger.addHandler(self.handler)

        dataframe_query_context_logger = PySparkLogger.getLogger("DataFrameQueryContextLogger")
        dataframe_query_context_logger.setLevel(logging.INFO)
        dataframe_query_context_logger.addHandler(self.handler)

    def test_log_structure(self):
        self.logger.info("Test logging structure")
        log_json = json.loads(self.handler.stream.getvalue().strip())
        keys = ["ts", "level", "logger", "msg", "context"]
        for key in keys:
            self.assertTrue(key in log_json)

    def test_log_info(self):
        self.logger.info("This is an info log", user="test_user_info", action="test_action_info")
        log_json = json.loads(self.handler.stream.getvalue().strip())

        self.assertEqual(log_json["msg"], "This is an info log")
        self.assertEqual(
            log_json["context"], {"action": "test_action_info", "user": "test_user_info"}
        )
        self.assertTrue("exception" not in log_json)

    def test_log_info_with_exception(self):
        # SPARK-51274: PySparkLogger should respect the expected keyword arguments
        self.logger.info(
            "This is an info log", exc_info=True, user="test_user_info", action="test_action_info"
        )
        log_json = json.loads(self.handler.stream.getvalue().strip())

        self.assertEqual(log_json["msg"], "This is an info log")
        self.assertEqual(
            log_json["context"], {"action": "test_action_info", "user": "test_user_info"}
        )
        self.assertTrue("exception" in log_json)
        self.assertTrue("class" in log_json["exception"])
        self.assertTrue("msg" in log_json["exception"])
        self.assertTrue("stacktrace" in log_json["exception"])

    def test_log_warning(self):
        self.logger.warning("This is an warn log", user="test_user_warn", action="test_action_warn")
        log_json = json.loads(self.handler.stream.getvalue().strip())

        self.assertEqual(log_json["msg"], "This is an warn log")
        self.assertEqual(
            log_json["context"], {"action": "test_action_warn", "user": "test_user_warn"}
        )
        self.assertTrue("exception" not in log_json)

    def test_log_error(self):
        self.logger.error(
            "This is an error log", user="test_user_error", action="test_action_error"
        )
        log_json = json.loads(self.handler.stream.getvalue().strip())

        self.assertEqual(log_json["msg"], "This is an error log")
        self.assertEqual(
            log_json["context"], {"action": "test_action_error", "user": "test_user_error"}
        )
        self.assertTrue("exception" not in log_json)

    def test_log_exception(self):
        self.logger.exception(
            "This is an exception log", user="test_user_exception", action="test_action_exception"
        )
        log_json = json.loads(self.handler.stream.getvalue().strip())

        self.assertEqual(log_json["msg"], "This is an exception log")
        self.assertEqual(
            log_json["context"], {"action": "test_action_exception", "user": "test_user_exception"}
        )
        self.assertTrue("exception" in log_json)
        self.assertTrue("class" in log_json["exception"])
        self.assertTrue("msg" in log_json["exception"])
        self.assertTrue("stacktrace" in log_json["exception"])

    def test_dataframe_query_context_logger(self):
        with self.sql_conf({"spark.sql.ansi.enabled": True}):
            df = self.spark.range(10)

            with self.assertRaises(ArithmeticException):
                df.withColumn("div_zero", df.id / 0).collect()
            log_json = json.loads(self.handler.stream.getvalue().strip())

            err_msg = "[DIVIDE_BY_ZERO] Division by zero."
            self.assertTrue(err_msg in log_json["msg"])
            self.assertTrue(err_msg in log_json["exception"]["msg"])
            self.assertEqual(log_json["context"]["fragment"], "__truediv__")
            self.assertEqual(log_json["context"]["errorClass"], "DIVIDE_BY_ZERO")
            # Only the class name is different between classic and connect.
            # Py4JJavaError for classic, _MultiThreadedRendezvous for connect
            self.assertTrue(
                log_json["exception"]["class"] in ("Py4JJavaError", "_MultiThreadedRendezvous")
            )
            stacktrace = log_json["exception"]["stacktrace"][0].keys()
            self.assertTrue("class" in stacktrace)
            self.assertTrue("method" in stacktrace)
            self.assertTrue("file" in stacktrace)
            self.assertTrue("line" in stacktrace)

    def test_log_exception_with_stacktrace(self):
        try:
            raise ValueError("Test Exception")
        except ValueError:
            self.logger.exception("Exception occurred", user="test_user_stacktrace")

        log_json = json.loads(self.handler.stream.getvalue().strip())

        self.assertEqual(log_json["msg"], "Exception occurred")
        self.assertEqual(log_json["context"], {"user": "test_user_stacktrace"})
        self.assertTrue("exception" in log_json)
        self.assertTrue("class" in log_json["exception"])
        self.assertTrue("msg" in log_json["exception"])
        self.assertTrue("stacktrace" in log_json["exception"])
        self.assertIsInstance(log_json["exception"]["stacktrace"], list)

        # Check the structure of "stacktrace"
        for frame in log_json["exception"]["stacktrace"]:
            self.assertTrue("class" in frame)
            self.assertTrue("method" in frame)
            self.assertTrue("file" in frame)
            self.assertTrue("line" in frame)

    def test_apply_schema(self):
        with tempfile.TemporaryDirectory(prefix="test_apply_schema") as d:
            logfile = f"{d}/log"
            handler = logging.FileHandler(logfile)
            try:
                self.logger.addHandler(handler)

                self.logger.info("Test logging structure.")
                try:
                    1 / 0
                except ZeroDivisionError:
                    self.logger.exception("Exception occurred.")

                assertDataFrameEqual(
                    self.spark.read.format("json")
                    .schema(SPARK_LOG_SCHEMA)
                    .load(logfile)
                    .select(
                        sf.col("ts").isNotNull().alias("ts_is_not_null"),
                        "level",
                        "msg",
                        sf.col("exception.class").alias("exception_class"),
                        sf.col("exception.msg").alias("exception_msg"),
                        sf.col("exception.stacktrace.class").alias("exception_stacktrace_class"),
                        sf.col("exception.stacktrace.method").alias("exception_stacktrace_method"),
                        sf.col("exception.stacktrace.file").alias("exception_stacktrace_file"),
                    ),
                    [
                        Row(
                            ts_is_not_null=True,
                            level="INFO",
                            msg="Test logging structure.",
                            exception_class=None,
                            exception_msg=None,
                            exception_stacktrace_class=None,
                            exception_stacktrace_method=None,
                            exception_stacktrace_file=None,
                        ),
                        Row(
                            ts_is_not_null=True,
                            level="ERROR",
                            msg="Exception occurred.",
                            exception_class="ZeroDivisionError",
                            exception_msg="division by zero",
                            exception_stacktrace_class=[None],
                            exception_stacktrace_method=["test_apply_schema"],
                            exception_stacktrace_file=[__file__],
                        ),
                    ],
                )
            finally:
                self.logger.removeHandler(handler)


class LoggerTests(LoggerTestsMixin, ReusedSQLTestCase):
    pass


if __name__ == "__main__":
    from pyspark.testing import main

    main()



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pandas as pd

from pyspark import pandas as ps
from pyspark.pandas.config import set_option, reset_option
from pyspark.testing.pandasutils import PandasOnSparkTestCase
from pyspark.testing.sqlutils import SQLTestUtils


class GroupByDiffMixin:
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        set_option("compute.ops_on_diff_frames", True)

    @classmethod
    def tearDownClass(cls):
        reset_option("compute.ops_on_diff_frames")
        super().tearDownClass()

    def test_diff(self):
        pdf = pd.DataFrame(
            {
                "a": [1, 2, 3, 4, 5, 6] * 3,
                "b": [1, 1, 2, 3, 5, 8] * 3,
                "c": [1, 4, 9, 16, 25, 36] * 3,
            }
        )
        pkey = pd.Series([1, 1, 2, 3, 5, 8] * 3)
        psdf = ps.from_pandas(pdf)
        kkey = ps.from_pandas(pkey)

        self.assert_eq(
            psdf.groupby(kkey).diff().sort_index(), pdf.groupby(pkey).diff().sort_index()
        )
        self.assert_eq(
            psdf.groupby(kkey)["a"].diff().sort_index(), pdf.groupby(pkey)["a"].diff().sort_index()
        )
        self.assert_eq(
            psdf.groupby(kkey)[["a"]].diff().sort_index(),
            pdf.groupby(pkey)[["a"]].diff().sort_index(),
        )

        self.assert_eq(psdf.groupby(kkey).diff().sum(), pdf.groupby(pkey).diff().sum().astype(int))
        self.assert_eq(psdf.groupby(kkey)["a"].diff().sum(), pdf.groupby(pkey)["a"].diff().sum())


class GroupByDiffTests(
    GroupByDiffMixin,
    PandasOnSparkTestCase,
    SQLTestUtils,
):
    pass


if __name__ == "__main__":
    from pyspark.testing import main

    main()



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import unittest

from numpy import array, random, exp, dot, all, mean, abs
from numpy import sum as array_sum

from pyspark import SparkContext
from pyspark.mllib.clustering import StreamingKMeans, StreamingKMeansModel
from pyspark.mllib.classification import StreamingLogisticRegressionWithSGD
from pyspark.mllib.linalg import Vectors
from pyspark.mllib.regression import LabeledPoint, StreamingLinearRegressionWithSGD
from pyspark.mllib.util import LinearDataGenerator
from pyspark.streaming import StreamingContext
from pyspark.testing.utils import eventually


class MLLibStreamingTestCase(unittest.TestCase):
    def setUp(self):
        self.sc = SparkContext("local[4]", "MLlib tests")
        self.ssc = StreamingContext(self.sc, 1.0)

    def tearDown(self):
        self.ssc.stop(False)
        self.sc.stop()


class StreamingKMeansTest(MLLibStreamingTestCase):
    def test_model_params(self):
        """Test that the model params are set correctly"""
        stkm = StreamingKMeans()
        stkm.setK(5).setDecayFactor(0.0)
        self.assertEqual(stkm._k, 5)
        self.assertEqual(stkm._decayFactor, 0.0)

        # Model not set yet.
        self.assertIsNone(stkm.latestModel())
        self.assertRaises(ValueError, stkm.trainOn, [0.0, 1.0])

        stkm.setInitialCenters(centers=[[0.0, 0.0], [1.0, 1.0]], weights=[1.0, 1.0])
        self.assertEqual(stkm.latestModel().centers, [[0.0, 0.0], [1.0, 1.0]])
        self.assertEqual(stkm.latestModel().clusterWeights, [1.0, 1.0])

    def test_accuracy_for_single_center(self):
        """Test that parameters obtained are correct for a single center."""
        centers, batches = self.streamingKMeansDataGenerator(
            batches=5, numPoints=5, k=1, d=5, r=0.1, seed=0
        )
        stkm = StreamingKMeans(1)
        stkm.setInitialCenters([[0.0, 0.0, 0.0, 0.0, 0.0]], [0.0])
        input_stream = self.ssc.queueStream([self.sc.parallelize(batch, 1) for batch in batches])
        stkm.trainOn(input_stream)

        self.ssc.start()

        def condition():
            self.assertEqual(stkm.latestModel().clusterWeights, [25.0])
            return True

        eventually(catch_assertions=True)(condition)()

        realCenters = array_sum(array(centers), axis=0)
        for i in range(5):
            modelCenters = stkm.latestModel().centers[0][i]
            self.assertAlmostEqual(centers[0][i], modelCenters, 1)
            self.assertAlmostEqual(realCenters[i], modelCenters, 1)

    def streamingKMeansDataGenerator(self, batches, numPoints, k, d, r, seed, centers=None):
        rng = random.RandomState(seed)

        # Generate centers.
        centers = [rng.randn(d) for i in range(k)]

        return centers, [
            [Vectors.dense(centers[j % k] + r * rng.randn(d)) for j in range(numPoints)]
            for i in range(batches)
        ]

    def test_trainOn_model(self):
        """Test the model on toy data with four clusters."""
        stkm = StreamingKMeans()
        initCenters = [[1.0, 1.0], [-1.0, 1.0], [-1.0, -1.0], [1.0, -1.0]]
        stkm.setInitialCenters(centers=initCenters, weights=[1.0, 1.0, 1.0, 1.0])

        # Create a toy dataset by setting a tiny offset for each point.
        offsets = [[0, 0.1], [0, -0.1], [0.1, 0], [-0.1, 0]]
        batches = []
        for offset in offsets:
            batches.append(
                [[offset[0] + center[0], offset[1] + center[1]] for center in initCenters]
            )

        batches = [self.sc.parallelize(batch, 1) for batch in batches]
        input_stream = self.ssc.queueStream(batches)
        stkm.trainOn(input_stream)
        self.ssc.start()

        # Give enough time to train the model.
        def condition():
            finalModel = stkm.latestModel()
            self.assertTrue(all(finalModel.centers == array(initCenters)))
            self.assertEqual(finalModel.clusterWeights, [5.0, 5.0, 5.0, 5.0])
            return True

        eventually(timeout=90, catch_assertions=True)(condition)()

    def test_predictOn_model(self):
        """Test that the model predicts correctly on toy data."""
        stkm = StreamingKMeans()
        stkm._model = StreamingKMeansModel(
            clusterCenters=[[1.0, 1.0], [-1.0, 1.0], [-1.0, -1.0], [1.0, -1.0]],
            clusterWeights=[1.0, 1.0, 1.0, 1.0],
        )

        predict_data = [[[1.5, 1.5]], [[-1.5, 1.5]], [[-1.5, -1.5]], [[1.5, -1.5]]]
        predict_data = [self.sc.parallelize(batch, 1) for batch in predict_data]
        predict_stream = self.ssc.queueStream(predict_data)
        predict_val = stkm.predictOn(predict_stream)

        result = []

        def update(rdd):
            rdd_collect = rdd.collect()
            if rdd_collect:
                result.append(rdd_collect)

        predict_val.foreachRDD(update)
        self.ssc.start()

        def condition():
            self.assertEqual(result, [[0], [1], [2], [3]])
            return True

        eventually(catch_assertions=True)(condition)()

    @unittest.skip("SPARK-10086: Flaky StreamingKMeans test in PySpark")
    def test_trainOn_predictOn(self):
        """Test that prediction happens on the updated model."""
        stkm = StreamingKMeans(decayFactor=0.0, k=2)
        stkm.setInitialCenters([[0.0], [1.0]], [1.0, 1.0])

        # Since decay factor is set to zero, once the first batch
        # is passed the clusterCenters are updated to [-0.5, 0.7]
        # which causes 0.2 & 0.3 to be classified as 1, even though the
        # classification based in the initial model would have been 0
        # proving that the model is updated.
        batches = [[[-0.5], [0.6], [0.8]], [[0.2], [-0.1], [0.3]]]
        batches = [self.sc.parallelize(batch) for batch in batches]
        input_stream = self.ssc.queueStream(batches)
        predict_results = []

        def collect(rdd):
            rdd_collect = rdd.collect()
            if rdd_collect:
                predict_results.append(rdd_collect)

        stkm.trainOn(input_stream)
        predict_stream = stkm.predictOn(input_stream)
        predict_stream.foreachRDD(collect)

        self.ssc.start()

        def condition():
            self.assertEqual(predict_results, [[0, 1, 1], [1, 0, 1]])
            return True

        eventually(catch_assertions=True)(condition)()


class StreamingLogisticRegressionWithSGDTests(MLLibStreamingTestCase):
    @staticmethod
    def generateLogisticInput(offset, scale, nPoints, seed):
        """
        Generate 1 / (1 + exp(-x * scale + offset))

        where,
        x is randomly distributed and the threshold
        and labels for each sample in x is obtained from a random uniform
        distribution.
        """
        rng = random.RandomState(seed)
        x = rng.randn(nPoints)
        sigmoid = 1.0 / (1 + exp(-(dot(x, scale) + offset)))
        y_p = rng.rand(nPoints)
        cut_off = y_p <= sigmoid
        y_p[cut_off] = 1.0
        y_p[~cut_off] = 0.0
        return [LabeledPoint(y_p[i], Vectors.dense([x[i]])) for i in range(nPoints)]

    def test_parameter_accuracy(self):
        """
        Test that the final value of weights is close to the desired value.
        """
        input_batches = [
            self.sc.parallelize(self.generateLogisticInput(0, 1.5, 100, 42 + i)) for i in range(20)
        ]
        input_stream = self.ssc.queueStream(input_batches)

        slr = StreamingLogisticRegressionWithSGD(stepSize=0.2, numIterations=25)
        slr.setInitialWeights([0.0])
        slr.trainOn(input_stream)

        self.ssc.start()

        def condition():
            rel = (1.5 - slr.latestModel().weights.array[0]) / 1.5
            self.assertAlmostEqual(rel, 0.1, 1)
            return True

        eventually(timeout=120.0, catch_assertions=True)(condition)()

    def test_convergence(self):
        """
        Test that weights converge to the required value on toy data.
        """
        input_batches = [
            self.sc.parallelize(self.generateLogisticInput(0, 1.5, 100, 42 + i)) for i in range(20)
        ]
        input_stream = self.ssc.queueStream(input_batches)
        models = []

        slr = StreamingLogisticRegressionWithSGD(stepSize=0.2, numIterations=25)
        slr.setInitialWeights([0.0])
        slr.trainOn(input_stream)

        def update_models(x):
            if not x.isEmpty():
                models.append(slr.latestModel().weights[0])

        input_stream.foreachRDD(update_models)

        self.ssc.start()

        def condition():
            self.assertEqual(len(models), len(input_batches))
            return True

        # We want all batches to finish for this test.
        eventually(timeout=120, catch_assertions=True)(condition)()

        t_models = array(models)
        diff = t_models[1:] - t_models[:-1]
        # Test that weights improve with a small tolerance
        self.assertTrue(all(diff >= -0.1))
        self.assertTrue(array_sum(diff > 0) > 1)

    @staticmethod
    def calculate_accuracy_error(true, predicted):
        return sum(abs(array(true) - array(predicted))) / len(true)

    def test_predictions(self):
        """Test predicted values on a toy model."""
        input_batches = []
        for i in range(20):
            batch = self.sc.parallelize(self.generateLogisticInput(0, 1.5, 100, 42 + i))
            input_batches.append(batch.map(lambda x: (x.label, x.features)))
        input_stream = self.ssc.queueStream(input_batches)

        slr = StreamingLogisticRegressionWithSGD(stepSize=0.2, numIterations=25)
        slr.setInitialWeights([1.5])
        predict_stream = slr.predictOnValues(input_stream)
        true_predicted = []
        predict_stream.foreachRDD(lambda x: true_predicted.append(x.collect()))
        self.ssc.start()

        def condition():
            self.assertEqual(len(true_predicted), len(input_batches))
            return True

        eventually(catch_assertions=True)(condition)()

        # Test that the accuracy error is no more than 0.4 on each batch.
        for batch in true_predicted:
            true, predicted = zip(*batch)
            self.assertTrue(self.calculate_accuracy_error(true, predicted) < 0.4)

    @unittest.skipIf(
        "COVERAGE_PROCESS_START" in os.environ, "Flaky with coverage enabled, skipping for now."
    )
    def test_training_and_prediction(self):
        """Test that the model improves on toy data with no. of batches"""
        input_batches = [
            self.sc.parallelize(self.generateLogisticInput(0, 1.5, 100, 42 + i)) for i in range(40)
        ]
        predict_batches = [b.map(lambda lp: (lp.label, lp.features)) for b in input_batches]

        slr = StreamingLogisticRegressionWithSGD(stepSize=0.01, numIterations=25)
        slr.setInitialWeights([-0.1])
        errors = []

        def collect_errors(rdd):
            true, predicted = zip(*rdd.collect())
            errors.append(self.calculate_accuracy_error(true, predicted))

        input_stream = self.ssc.queueStream(input_batches)
        predict_stream = self.ssc.queueStream(predict_batches)
        slr.trainOn(input_stream)
        ps = slr.predictOnValues(predict_stream)
        ps.foreachRDD(lambda x: collect_errors(x))

        self.ssc.start()

        def condition():
            # Test that the improvement in error is > 0.3
            if len(errors) == len(predict_batches):
                self.assertGreater(errors[1] - errors[-1], 0.3)
            if len(errors) >= 3 and errors[1] - errors[-1] > 0.3:
                return True
            return "Latest errors: " + ", ".join(map(lambda x: str(x), errors))

        eventually(timeout=180.0)(condition)()


class StreamingLinearRegressionWithTests(MLLibStreamingTestCase):
    def assertArrayAlmostEqual(self, array1, array2, dec):
        for i, j in array1, array2:
            self.assertAlmostEqual(i, j, dec)

    def test_parameter_accuracy(self):
        """Test that coefs are predicted accurately by fitting on toy data."""

        # Test that fitting (10*X1 + 10*X2), (X1, X2) gives coefficients
        # (10, 10)
        slr = StreamingLinearRegressionWithSGD(stepSize=0.2, numIterations=25)
        slr.setInitialWeights([0.0, 0.0])
        xMean = [0.0, 0.0]
        xVariance = [1.0 / 3.0, 1.0 / 3.0]

        # Create ten batches with 100 sample points in each.
        batches = []
        for i in range(10):
            batch = LinearDataGenerator.generateLinearInput(
                0.0, [10.0, 10.0], xMean, xVariance, 100, 42 + i, 0.1
            )
            batches.append(self.sc.parallelize(batch))

        input_stream = self.ssc.queueStream(batches)
        slr.trainOn(input_stream)
        self.ssc.start()

        def condition():
            self.assertArrayAlmostEqual(slr.latestModel().weights.array, [10.0, 10.0], 1)
            self.assertAlmostEqual(slr.latestModel().intercept, 0.0, 1)
            return True

        eventually(catch_assertions=True)(condition)()

    def test_parameter_convergence(self):
        """Test that the model parameters improve with streaming data."""
        slr = StreamingLinearRegressionWithSGD(stepSize=0.2, numIterations=25)
        slr.setInitialWeights([0.0])

        # Create ten batches with 100 sample points in each.
        batches = []
        for i in range(10):
            batch = LinearDataGenerator.generateLinearInput(
                0.0, [10.0], [0.0], [1.0 / 3.0], 100, 42 + i, 0.1
            )
            batches.append(self.sc.parallelize(batch))

        model_weights = []
        input_stream = self.ssc.queueStream(batches)
        input_stream.foreachRDD(lambda x: model_weights.append(slr.latestModel().weights[0]))
        slr.trainOn(input_stream)
        self.ssc.start()

        def condition():
            self.assertEqual(len(model_weights), len(batches))
            return True

        # We want all batches to finish for this test.
        eventually(timeout=90, catch_assertions=True)(condition)()

        w = array(model_weights)
        diff = w[1:] - w[:-1]
        self.assertTrue(all(diff >= -0.1))

    def test_prediction(self):
        """Test prediction on a model with weights already set."""
        # Create a model with initial Weights equal to coefs
        slr = StreamingLinearRegressionWithSGD(stepSize=0.2, numIterations=25)
        slr.setInitialWeights([10.0, 10.0])

        # Create ten batches with 100 sample points in each.
        batches = []
        for i in range(10):
            batch = LinearDataGenerator.generateLinearInput(
                0.0, [10.0, 10.0], [0.0, 0.0], [1.0 / 3.0, 1.0 / 3.0], 100, 42 + i, 0.1
            )
            batches.append(self.sc.parallelize(batch).map(lambda lp: (lp.label, lp.features)))

        input_stream = self.ssc.queueStream(batches)
        output_stream = slr.predictOnValues(input_stream)
        samples = []
        output_stream.foreachRDD(lambda x: samples.append(x.collect()))

        self.ssc.start()

        def condition():
            self.assertEqual(len(samples), len(batches))
            return True

        # We want all batches to finish for this test.
        eventually(catch_assertions=True)(condition)()

        # Test that mean absolute error on each batch is less than 0.1
        for batch in samples:
            true, predicted = zip(*batch)
            self.assertTrue(mean(abs(array(true) - array(predicted))) < 0.1)

    @unittest.skipIf(
        "COVERAGE_PROCESS_START" in os.environ, "Flaky with coverage enabled, skipping for now."
    )
    def test_train_prediction(self):
        """Test that error on test data improves as model is trained."""
        slr = StreamingLinearRegressionWithSGD(stepSize=0.2, numIterations=25)
        slr.setInitialWeights([0.0])

        # Create fifteen batches with 100 sample points in each.
        batches = []
        for i in range(15):
            batch = LinearDataGenerator.generateLinearInput(
                0.0, [10.0], [0.0], [1.0 / 3.0], 100, 42 + i, 0.1
            )
            batches.append(self.sc.parallelize(batch))

        predict_batches = [b.map(lambda lp: (lp.label, lp.features)) for b in batches]
        errors = []

        def func(rdd):
            true, predicted = zip(*rdd.collect())
            errors.append(mean(abs(true) - abs(predicted)))

        input_stream = self.ssc.queueStream(batches)
        output_stream = self.ssc.queueStream(predict_batches)
        slr.trainOn(input_stream)
        output_stream = slr.predictOnValues(output_stream)
        output_stream.foreachRDD(func)
        self.ssc.start()

        def condition():
            if len(errors) == len(predict_batches):
                self.assertGreater(errors[1] - errors[-1], 2)
            if len(errors) >= 3 and errors[1] - errors[-1] > 2:
                return True
            return "Latest errors: " + ", ".join(map(lambda x: str(x), errors))

        eventually(timeout=180.0)(condition)()


if __name__ == "__main__":
    from pyspark.testing import main

    main()



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Util functions for workers.
"""

from contextlib import contextmanager
import importlib
from inspect import currentframe, getframeinfo
import os
import sys
from typing import Any, Generator, IO, Optional
import warnings

if "SPARK_TESTING" in os.environ:
    assert os.environ.get("SPARK_PYTHON_RUNTIME") == "PYTHON_WORKER", (
        "This module can only be imported in python woker"
    )

# 'resource' is a Unix specific module.
has_resource_module = True
try:
    import resource
except ImportError:
    has_resource_module = False

from pyspark.accumulators import _accumulatorRegistry
from pyspark.util import is_remote_only
from pyspark.errors import PySparkRuntimeError
from pyspark.util import local_connect_and_auth
from pyspark.serializers import (
    read_bool,
    read_int,
    read_long,
    write_int,
    FramedSerializer,
    UTF8Deserializer,
    CPickleSerializer,
)

pickleSer = CPickleSerializer()
utf8_deserializer = UTF8Deserializer()


def add_path(path: str) -> bool:
    # worker can be used, so do not add path multiple times
    if path not in sys.path:
        # overwrite system packages
        sys.path.insert(1, path)
        return True
    return False


def read_command(serializer: FramedSerializer, file: IO) -> Any:
    if not is_remote_only():
        from pyspark.core.broadcast import Broadcast

    command = serializer._read_with_length(file)
    if not is_remote_only() and isinstance(command, Broadcast):
        command = serializer.loads(command.value)
    return command


def check_python_version(infile: IO) -> None:
    """
    Check the Python version between the running process and the one used to serialize the command.
    """
    version = utf8_deserializer.loads(infile)
    worker_version = "%d.%d" % sys.version_info[:2]
    if version != worker_version:
        raise PySparkRuntimeError(
            errorClass="PYTHON_VERSION_MISMATCH",
            messageParameters={
                "worker_version": worker_version,
                "driver_version": str(version),
            },
        )


def setup_memory_limits(memory_limit_mb: int) -> None:
    """
    Sets up the memory limits.

    If memory_limit_mb > 0 and `resource` module is available, sets the memory limit.
    Windows does not support resource limiting and actual resource is not limited on MacOS.
    """
    if memory_limit_mb > 0 and has_resource_module:
        total_memory = resource.RLIMIT_AS
        try:
            soft_limit, hard_limit = resource.getrlimit(total_memory)
            msg = "Current mem limits: {0} of max {1}\n".format(soft_limit, hard_limit)
            print(msg, file=sys.stderr)

            # convert to bytes
            new_limit = memory_limit_mb * 1024 * 1024

            if soft_limit == resource.RLIM_INFINITY or new_limit < soft_limit:
                msg = "Setting mem limits to {0} of max {1}\n".format(new_limit, new_limit)
                print(msg, file=sys.stderr)
                resource.setrlimit(total_memory, (new_limit, new_limit))

        except (resource.error, OSError, ValueError) as e:
            # not all systems support resource limits, so warn instead of failing
            current = currentframe()
            lineno = getframeinfo(current).lineno + 1 if current is not None else 0
            if "__file__" in globals():
                print(
                    warnings.formatwarning(
                        "Failed to set memory limit: {0}".format(e),
                        ResourceWarning,
                        __file__,
                        lineno,
                    ),
                    file=sys.stderr,
                )


def setup_spark_files(infile: IO) -> None:
    """
    Set up Spark files, archives, and pyfiles.
    """
    # fetch name of workdir
    spark_files_dir = utf8_deserializer.loads(infile)

    if not is_remote_only():
        from pyspark.core.files import SparkFiles

        SparkFiles._root_directory = spark_files_dir
        SparkFiles._is_running_on_worker = True

    # fetch names of includes (*.zip and *.egg files) and construct PYTHONPATH
    path_changed = add_path(spark_files_dir)  # *.py files that were added will be copied here
    num_python_includes = read_int(infile)
    for _ in range(num_python_includes):
        filename = utf8_deserializer.loads(infile)
        path_changed = add_path(os.path.join(spark_files_dir, filename)) or path_changed

    if path_changed:
        importlib.invalidate_caches()


def setup_broadcasts(infile: IO) -> None:
    """
    Set up broadcasted variables.
    """
    if not is_remote_only():
        from pyspark.core.broadcast import Broadcast, _broadcastRegistry

    # fetch names and values of broadcast variables
    needs_broadcast_decryption_server = read_bool(infile)
    num_broadcast_variables = read_int(infile)
    if needs_broadcast_decryption_server:
        # read the decrypted data from a server in the jvm
        conn_info = read_int(infile)
        auth_secret = None
        if conn_info == -1:
            conn_info = utf8_deserializer.loads(infile)
        else:
            auth_secret = utf8_deserializer.loads(infile)
        broadcast_sock_file, _ = local_connect_and_auth(conn_info, auth_secret)

    for _ in range(num_broadcast_variables):
        bid = read_long(infile)
        if bid >= 0:
            if needs_broadcast_decryption_server:
                read_bid = read_long(broadcast_sock_file)
                assert read_bid == bid
                _broadcastRegistry[bid] = Broadcast(sock_file=broadcast_sock_file)
            else:
                path = utf8_deserializer.loads(infile)
                _broadcastRegistry[bid] = Broadcast(path=path)

        else:
            bid = -bid - 1
            _broadcastRegistry.pop(bid)

    if needs_broadcast_decryption_server:
        broadcast_sock_file.write(b"1")
        broadcast_sock_file.close()


@contextmanager
def get_sock_file_to_executor(timeout: Optional[int] = -1) -> Generator[IO, None, None]:
    # Read information about how to connect back to the JVM from the environment.
    conn_info = os.environ.get(
        "PYTHON_WORKER_FACTORY_SOCK_PATH", int(os.environ.get("PYTHON_WORKER_FACTORY_PORT", -1))
    )
    auth_secret = os.environ.get("PYTHON_WORKER_FACTORY_SECRET")
    sock_file, sock = local_connect_and_auth(conn_info, auth_secret)
    if timeout is None or timeout > 0:
        sock.settimeout(timeout)
    # TODO: Remove the following two lines and use `Process.pid()` when we drop JDK 8.
    write_int(os.getpid(), sock_file)
    sock_file.flush()
    try:
        yield sock_file
    finally:
        sock_file.close()


def send_accumulator_updates(outfile: IO) -> None:
    """
    Send the accumulator updates back to JVM.
    """
    write_int(len(_accumulatorRegistry), outfile)
    for aid, accum in _accumulatorRegistry.items():
        pickleSer._write_with_length((aid, accum._value), outfile)


class Conf:
    def __init__(self, infile: Optional[IO] = None) -> None:
        self._conf: dict[str, Any] = {}
        if infile is not None:
            self.load(infile)

    def load(self, infile: IO) -> None:
        num_conf = read_int(infile)
        # We do a sanity check here to reduce the possibility to stuck indefinitely
        # due to an invalid messsage. If the numer of configurations is obviously
        # wrong, we just raise an error directly.
        # We hand-pick the configurations to send to the worker so the number should
        # be very small (less than 100).
        if num_conf < 0 or num_conf > 10000:
            raise PySparkRuntimeError(
                errorClass="PROTOCOL_ERROR",
                messageParameters={
                    "failure": f"Invalid number of configurations: {num_conf}",
                },
            )
        for _ in range(num_conf):
            k = utf8_deserializer.loads(infile)
            v = utf8_deserializer.loads(infile)
            self._conf[k] = v

    def get(self, key: str, default: Any = "", *, lower_str: bool = True) -> Any:
        val = self._conf.get(key, default)
        if isinstance(val, str) and lower_str:
            return val.lower()
        return val
